"""
Object describing the model values, i.e. this uniquely
describes the set of scaling relations present.
"""

import attr
import numpy as np
import yaml
import json

from typing import Dict, Optional, Hashable
from pathlib import Path


@attr.s
class ModelValues(object):
    """
    Set of model values for a given scaling relation.

    Parameters
    ----------

    model_values, Dict[Hashable, Dict[str, Optional[np.array]]]
        Model values for this given scaling relation. Must have the
        following structure: ``{unique_identifier: {"independent":
        np.array, "dependent": np.array, "dependent_errors": Optional[
        np.array]}}``, with the ``unique_identifiers`` the same as
        those that were specified within the :class:`ModelParameters`
        object. The dependent errors are optional, but if present
        must be of the same length as the dependent variables. Both
        independent and dependent are required for every model value
        that is present. If a unique identifier is present in the
        model parameters, but not in the values, it is simply left
        out of the emulation step.

    Raises
    ------

    AttributeError
        If the model values do not conform to the required
        specification.
    """

    model_values: Dict[Hashable, Dict[str, Optional[np.array]]] = attr.ib()

    @model_values.validator
    def _check_model_values(self, attribute, value):
        """
        Checks the model values to ensure that all
        required fields are present.
        """

        for unique_id, scaling_relation in value.items():
            if not (
                "independent" in scaling_relation.keys()
                or "dependent" in scaling_relation.keys()
            ):
                raise AttributeError(
                    "independent and dependent must be available for all models, "
                    f"missing for model {unique_id}."
                )

            if "dependent_error" in scaling_relation.keys():
                if not len(scaling_relation["dependent_error"]) == len(
                    scaling_relation["dependent"]
                ):
                    raise AttributeError(
                        "dependent_error present but not the same length as "
                        f"dependent for model {unique_id}."
                    )

            acceptable_keys = ["independent", "dependent", "dependent_error"]

            for key in scaling_relation.keys():
                if key not in acceptable_keys:
                    raise AttributeError(
                        f"Invalid key {key} in model values container. Choose from "
                        f"one of {acceptable_keys}."
                    )

        return

    def items(self):
        return self.model_values.items()

    def keys(self):
        return self.model_values.keys()

    def values(self):
        return self.model_values.values()

    def __getitem__(self, key):
        return self.model_values[key]

    def __len__(self):
        return len(self.model_values)

    @property
    def number_of_variables(self) -> int:
        """
        Total number of variables present in all models.
        """

        total = sum([len(x["independent"]) for x in self.model_values.values()])

        return total

    def to_yaml(self, filename: Path):
        """
        Write the model values to a YAML file.

        Parameters
        ----------

        filename: Path
            The path to write the file to. This should be a `Path` object,
            but if it is a string it will be automatically converted.

        """

        filename = Path(filename)

        # We must first construct a version of `model_values` that uses lists instead
        # of numpy arrays - yaml doesn't like that!

        listify = lambda x: {
            k: (v.tolist() if isinstance(v, np.ndarray) else v) for k, v in x.items()
        }

        model_value_no_numpy = {
            uid: listify(model) for uid, model in self.model_values.items()
        }

        with open(filename, "w") as handle:
            yaml.dump(model_value_no_numpy, stream=handle)

        return

    @classmethod
    def from_yaml(cls, filename: Path) -> "ModelValues":
        """
        Generate an instance of :class:`ModelValues` from a YAML file,
        written to disk using `to_yaml`.

        Parameters
        ----------

        filename: Path
            The path to read the file from. This should be a `Path` object,
            but if it is a string it will be automatically converted.

        Returns
        -------

        model_values: ModelValues
            Instance of `ModelValues` restored from disk.

        """

        filename = Path(filename)

        with open(filename, "r") as handle:
            raw_values = dict(yaml.load(stream=handle, Loader=yaml.FullLoader))

        # Now need to re-numpify

        numpyify = lambda x: {
            k: (np.array(v) if isinstance(v, list) else v) for k, v in x.items()
        }

        model_value_numpy = {uid: numpyify(model) for uid, model in raw_values.items()}

        return cls(model_values=model_value_numpy)

    def to_json(self, filename: Path):
        """
        Write the model values to a JSON file. Preferred to YAML as this
        is much faster for large datsets.

        Parameters
        ----------

        filename: Path
            The path to write the file to. This should be a `Path` object,
            but if it is a string it will be automatically converted.

        """

        filename = Path(filename)

        # We must first construct a version of `model_values` that uses lists instead
        # of numpy arrays - yaml doesn't like that!

        listify = lambda x: {
            k: (v.tolist() if isinstance(v, np.ndarray) else v) for k, v in x.items()
        }

        model_value_no_numpy = {
            uid: listify(model) for uid, model in self.model_values.items()
        }

        with open(filename, "w") as handle:
            json.dump(model_value_no_numpy, handle)

        return

    @classmethod
    def from_json(cls, filename: Path) -> "ModelValues":
        """
        Generate an instance of :class:`ModelValues` from a JSON file,
        written to disk using ``to_json``.

        Parameters
        ----------

        filename: Path
            The path to read the file from. This should be a `Path` object,
            but if it is a string it will be automatically converted.

        Returns
        -------

        model_values: ModelValues
            Instance of ``ModelValues`` restored from disk.

        """

        filename = Path(filename)

        with open(filename, "r") as handle:
            raw_values = dict(json.load(handle))

        # Now need to re-numpify

        numpyify = lambda x: {
            k: (np.array(v) if isinstance(v, list) else v) for k, v in x.items()
        }

        model_value_numpy = {uid: numpyify(model) for uid, model in raw_values.items()}

        return cls(model_values=model_value_numpy)



"""
Generates a completely random design, by using the numpy random command.
"""

from swiftemulator.backend.model_specification import ModelSpecification
from swiftemulator.backend.model_parameters import ModelParameters
from swiftemulator.design.transform import transform_to_model_spec

from typing import Optional, Dict, Any, Optional

import numpy as np


def create_cube(
    model_specification: ModelSpecification,
    number_of_samples: int,
    prefix_unique_id: Optional[str] = None,
) -> ModelParameters:
    """
    Creates a random hypercube model design.

    Parameters
    ----------

    model_specification: ModelSpecification
        Model specification for which to create a latin hypercube
        from.

    number_of_samples: int
        The number of samples to draw; this will be the number
        of input simulations that you wish to create.

    prefix_unique_id: str, optional
        An optional prefix for the newly generated unique IDs.
        Defaults to no prefix.


    Returns
    -------

    model_parameters: ModelParameters
        A model values container with the prepared latin hypercube.
        Contains methods to visualise the output hypercube.


    Notes
    -----

    Uses numpy's random methods to generate a completely random (i.e.
    no guarantee of a nice even distribution) hypercube
    """

    samples = np.random.rand(
        number_of_samples, model_specification.number_of_parameters
    )

    return transform_to_model_spec(
        input_array=samples,
        model_specification=model_specification,
        prefix_unique_id=prefix_unique_id,
    )



"""
Transformer from an ND array to a model specification object.
"""

from swiftemulator.backend.model_parameters import ModelParameters
from swiftemulator.backend.model_specification import ModelSpecification

import numpy as np

from typing import Optional


def transform_to_model_spec(
    input_array: np.ndarray,
    model_specification: ModelSpecification,
    prefix_unique_id: Optional[str] = None,
) -> ModelParameters:
    """
    Transforms the input nd array (which is of shape n models
    by n parameters) to a model parameters object, by re-scaling the
    parameters according to the specification.

    Parameters
    ----------

    input_array: np.ndarray
        Input array, of shape (number of samples, number of parameters).

    model_specification: ModelSpecification
        Model specification used to rescale the array.

    prefix_unique_id: str, optional
        An optional prefix for the newly generated unique IDs.
        Defaults to no prefix.


    Returns
    -------

    model_parameters: ModelParameters
        A model values container with the re-scaled parameters and
        associated metadata.
    """

    transform = lambda i, l: float((i * (l[1] - l[0])) + l[0])
    prefix = prefix_unique_id if prefix_unique_id is not None else ""
    number_of_samples = len(input_array)

    model_parameters = {
        f"{prefix}{key}": {
            par: transform(input_array[key][i], model_specification.parameter_limits[i])
            for i, par in enumerate(model_specification.parameter_names)
        }
        for key in range(number_of_samples)
    }

    return ModelParameters(model_parameters=model_parameters)



"""
Generates a latin hypercube ``ModelValues`` container given
the ``ModelSpecification``. Uses :mod:``pyDOE``.
"""

from pyDOE import lhs

from swiftemulator.backend.model_specification import ModelSpecification
from swiftemulator.backend.model_parameters import ModelParameters
from swiftemulator.design.transform import transform_to_model_spec

from typing import Optional, Dict, Any, Optional
from contextlib import redirect_stdout
from os import devnull

import numpy as np


def create_hypercube(
    model_specification: ModelSpecification,
    number_of_samples: int,
    correlation_retries: Optional[int] = 32,
    prefix_unique_id: Optional[str] = None,
) -> ModelParameters:
    """
    Creates a Latin Hypercube model design.

    Parameters
    ----------

    model_specification: ModelSpecification
        Model specification for which to create a latin hypercube
        from.

    number_of_samples: int
        The number of samples to draw; this will be the number
        of input simulations that you wish to create.

    correlation_retries: int, optional
        Number of times to re-try creating a random hypercube, to
        minimize the correlation coefficient even further.
        Default: 32.

    prefix_unique_id: str, optional
        An optional prefix for the newly generated unique IDs.
        Defaults to no prefix.


    Returns
    -------

    model_parameters: ModelParameters
        A model values container with the prepared latin hypercube.
        Contains methods to visualise the output hypercube.


    Notes
    -----

    Uses :mod:``pyDOE``'s :func:`lhs` function, with the ``maximin``
    method.
    """

    samples = None
    corr = 1.0

    for _ in range(correlation_retries):
        # Reduce correlation as much as practical
        with redirect_stdout(open(devnull, "w")):
            new_samples = lhs(
                n=model_specification.number_of_parameters,
                samples=number_of_samples,
                criterion="maximin",
            )

        R = np.corrcoef(new_samples)
        min_corr = np.max(np.abs(R - np.eye(R.shape[0])))

        if min_corr <= corr:
            samples = new_samples
            corr = min_corr

    return transform_to_model_spec(
        input_array=samples,
        model_specification=model_specification,
        prefix_unique_id=prefix_unique_id,
    )



"""
Tests the models in mean_models.
"""

from swiftemulator.mean_models.offset import OffsetMeanModel
from swiftemulator.mean_models.linear import LinearMeanModel
from swiftemulator.mean_models.polynomial import PolynomialMeanModel

import numpy as np


def test_offset_mean_model():
    mean_model = OffsetMeanModel()

    mean_model.train(independent=np.random.rand(100, 23), dependent=np.arange(100))

    predicted = mean_model.predict(
        independent=np.random.rand(12, 23),
    )

    assert (predicted == 49.5).all()


def test_linear_mean_model():
    mean_model = LinearMeanModel(lasso_model_alpha=1.0)

    mean_model.train(independent=np.random.rand(100, 10), dependent=np.arange(100))

    mean_model.predict(
        independent=np.random.rand(12, 10),
    )

    george_model = mean_model.george_model

    george_model.get_value(np.random.rand(12, 10))


def test_polynomial_mean_model():
    mean_model = PolynomialMeanModel(degree=4)

    mean_model.train(independent=np.random.rand(100, 10), dependent=np.arange(100))

    mean_model.predict(
        independent=np.random.rand(12, 10),
    )

    george_model = mean_model.george_model

    george_model.get_value(np.random.rand(12, 10))



"""
Tests the emulator in multi_guassian_process.
"""

from swiftemulator.backend.emulator_generator import (
    ModelParameters,
    ModelSpecification,
    ModelValues,
)

from swiftemulator.emulators.multi_gaussian_process import (
    MultipleGaussianProcessEmulator,
)

import numpy as np


def test_basic_emulator_generator():
    """
    Basic emulator generator test. This should fall back to a single
    emulator.
    """

    model_spec = ModelSpecification(
        number_of_parameters=2,
        parameter_names=["x", "y"],
        parameter_limits=[[1.0, 3.0], [0.0, 1.0]],
    )

    my_model_parameters = {
        0: {
            "x": 1,
            "y": 1,
        },
        1: {
            "x": 2,
            "y": 1,
        },
        2: {"x": 3, "y": 1},
    }

    model_parameters = ModelParameters(model_parameters=my_model_parameters)

    input_model_values = {
        0: {
            "independent": np.arange(10),
            "dependent": np.random.rand(10),
            "dependent_error": np.random.rand(10),
        },
        1: {
            "independent": np.arange(10),
            "dependent": np.random.rand(10),
            "dependent_error": np.random.rand(10),
        },
        2: {
            "independent": np.arange(10),
            "dependent": np.random.rand(10),
        },
    }

    model_values = ModelValues(model_values=input_model_values)

    gpe = MultipleGaussianProcessEmulator()

    gpe.fit_model(
        model_specification=model_spec,
        model_parameters=model_parameters,
        model_values=model_values,
    )


def test_basic_emulator_generator_multiple():
    """
    Basic emulator generator test, with multiple trained emulators.
    """

    model_spec = ModelSpecification(
        number_of_parameters=2,
        parameter_names=["x", "y"],
        parameter_limits=[[1.0, 3.0], [0.0, 1.0]],
    )

    my_model_parameters = {
        0: {
            "x": 1,
            "y": 1,
        },
        1: {
            "x": 2,
            "y": 1,
        },
        2: {"x": 3, "y": 1},
    }

    model_parameters = ModelParameters(model_parameters=my_model_parameters)

    input_model_values = {
        0: {
            "independent": np.arange(10),
            "dependent": np.random.rand(10),
            "dependent_error": np.random.rand(10),
        },
        1: {
            "independent": np.arange(10),
            "dependent": np.random.rand(10),
            "dependent_error": np.random.rand(10),
        },
        2: {
            "independent": np.arange(10),
            "dependent": np.random.rand(10),
        },
    }

    model_values = ModelValues(model_values=input_model_values)

    gpe = MultipleGaussianProcessEmulator(independent_regions=[[None, 6], [4, None]])

    gpe.fit_model(
        model_specification=model_spec,
        model_parameters=model_parameters,
        model_values=model_values,
    )

    gpe.predict_values(np.array([0.2, 0.9, 9.9, 5.0]), {"x": 0.5, "y": 1})
    gpe.predict_values_no_error(np.array([0.2, 0.9, 9.9, 5.0]), {"x": 0.5, "y": 1})



"""
Mean models, usually used to create a basic parametrisation
of the space to enable the GPE to more efficiently model residuals.
"""

from swiftemulator.mean_models.base import MeanModel
from swiftemulator.mean_models.linear import LinearMeanModel
from swiftemulator.mean_models.offset import OffsetMeanModel
from swiftemulator.mean_models.polynomial import PolynomialMeanModel
from swiftemulator.mean_models.fixed import FixedMeanModel



"""
A basic linear model based on the scikit-learn multidimensional
mean model.
"""

import attr
import numpy as np
import sklearn.linear_model as lm

from .base import MeanModel

from typing import Optional, Union


@attr.s
class LinearMeanModel(MeanModel):
    """
    A linear mean model; fits a linear model to the
    multidimensional parameter space.

    Under the hood, this uses the ``sklearn.linear_model.lm``.

    Parameters
    ----------

    lasso_model_alpha: float, optional
        ``alpha`` for the Lasso model. If this is zero (the default)
        we fit a basic linear regression model is used for performance
        reasons.
    """

    lasso_model_alpha: float = attr.ib(default=0.0)

    model: Optional[Union[lm.LinearRegression, lm.Lasso]] = None

    def train(self, independent: np.ndarray, dependent: np.ndarray) -> None:
        """
        Train the model. See :class:`MeanModel` for more information.
        """

        self.model = (
            lm.LinearRegression(fit_intercept=True)
            if self.lasso_model_alpha == 0.0
            else lm.Lasso(alpha=self.lasso_model_alpha)
        )

        self.model.fit(independent, dependent)

        return

    def predict(self, independent: np.ndarray) -> np.ndarray:
        """
        Predict using the model. See :class:`MeanModel` for more information.
        """

        dependent = self.model.predict(independent)

        return dependent



"""
Tests for objects in ``swift-emulator/backend/model_specification.py``
"""

from swiftemulator.backend.model_parameters import ModelParameters
import pytest


def test_basic_model_parameters():
    """
    A basic test with a valid set of model parameters.
    """

    my_model_parameters = {
        0: {
            "x": 1,
            "y": 1,
        },
        1: {
            "x": 2,
            "y": 1,
        },
    }

    _ = ModelParameters(model_parameters=my_model_parameters)

    return


def test_basic_model_parameters_raise():
    """
    A basic test with an invalid set of model parameters.
    """

    my_model_parameters = {
        0: {
            "x": 1,
            "y": 1,
        },
        1: {
            "x": 2,
            "z": 1,
        },
    }

    with pytest.raises(AttributeError):
        _ = ModelParameters(model_parameters=my_model_parameters)

    return


def test_find_closest_model():
    """
    A simple test to see if it finds the closest model
    """

    model_point = {"x": 0.4, "y": 0.9}

    my_model_parameters = {
        0: {
            "x": 1,
            "y": 1,
        },
        1: {
            "x": 2,
            "y": 1,
        },
        2: {
            "x": 0.45,
            "y": 0.91,
        },
    }

    my_model_parameters = ModelParameters(model_parameters=my_model_parameters)
    model, _ = my_model_parameters.find_closest_model(model_point)

    assert model[0] == 2

    model, _ = my_model_parameters.find_closest_model(
        model_point, number_of_close_models=2
    )

    assert model[0] == 2

    return



"""
Penalty functions and their specifications.
"""

import attr
import unyt
import numpy as np
import matplotlib.pyplot as plt

from velociraptor.observations import ObservationalData
from scipy.interpolate import interp1d

from swiftemulator.backend.model_values import ModelValues

from typing import Dict, Hashable, Optional, Union, Callable, List


@attr.s
class PenaltyCalculator(object):
    """
    Base class for the penalty functions.

    Extend this with your own, following the following pattern:

    1. Configuration parameters for the penalty function
       are taken as initialisation parameters to the class.
    2. The observational data set to be matched to is passed
       to register_observation.
    3. The penalty function is calculated using the penalty()
       method, taking the exact arguments that are taken here,
       for an individual model.

    Provided for convenience is penalties() which calculates the
    penalty function for all data in a `ModelValues` container.
    """

    observation: ObservationalData
    interpolator_values: interp1d
    interpolator_error: Optional[interp1d]

    log_independent: bool
    log_dependent: bool
    independent_units: unyt.unyt_quantity
    dependent_units: unyt.unyt_quantity

    def register_observation(
        self,
        observation: ObservationalData,
        log_independent: bool,
        log_dependent: bool,
        independent_units: unyt.unyt_quantity,
        dependent_units: unyt.unyt_quantity,
    ) -> None:
        """
        Registers the observation for use in `penalty` with the class.

        Parameters
        ----------

        observation: ObservationalData
            Instance of the velociraptor observational data used
            for comparisons.

        log_independent: bool
            Take the base-10 log of the independent data before comparison?

        log_dependent: bool
            Take the base-10 log of the dependent data before comparison?

        independent_units: unyt.unyt_quantity
            The units that the model was calculated in (independent)

        dependent_units: unyt.unyt_quantity
            The units that the model was calculated in (dependent)

        """

        self.observation = observation
        self.log_independent = log_independent
        self.log_dependent = log_dependent
        self.independent_units = independent_units
        self.dependent_units = dependent_units

        self.observation_interpolation()

        return

    def observation_interpolation(self):
        """
        Produces the interpolation for the internal observation.
        """

        x = self.observation.x.to(self.independent_units)
        y = self.observation.y.to(self.dependent_units)

        if self.log_independent:
            x = np.log10(x.value)

        if self.log_dependent:
            y = np.log10(y.value)

        fill_value = lambda x, y: (y[x.argmin()], y[x.argmax()])

        self.interpolator_values = interp1d(
            x=x,
            y=y,
            kind="linear",
            copy=False,
            bounds_error=False,
            fill_value=fill_value(x, y),
        )

        # Set observational error to None for now.
        self.interpolator_error = None

        return

    def penalty(
        self,
        independent: np.array,
        dependent: np.array,
        dependent_error: Optional[np.array] = None,
    ) -> List[float]:
        """
        Calculate the penalty function, relative to the observational
        data, for this model. It is highly recommended that you evaluate
        the model at the same independent variables as the observational
        data. The observational data is linearly interpolated to find a
        prediction at the independent variables that you provide.

        independent: np.array
            The independent data.

        dependent: np.array
            The dependent data for comparison.

        dependent_error: np.array, optional
            The dependent errors, for comparison.

        Returns
        -------

        penalty, List[float]
            The penalties for this model, between 0 and 1 each.
        """

        raise NotImplementedError

    def penalties(
        self, model: ModelValues, collate_with: Callable
    ) -> Dict[Hashable, float]:
        """
        Calculate the penalty function for all models in the
        model values container.

        It is highly recommended that you evaluate the model at the same
        independent variables as the observational data. The observational
        data is linearly interpolated to find a prediction at the independent
        variables that you provide.

        Parameters
        ----------

        model: ModelValues
            The set of model (values) to calculate the penalty
            function for.

        collate_with: Callable
            A function that takes a numpy array and returns the
            'global' penalty for a model given the input for all
            of the valid points in the array. Examples could be
            ``np.max``, ``np.mean``, ``np.median``.


        Returns
        -------

        penalties: Dict[Hashable, float]
            Penalty functions for each of the models, with the key
            being the unique identifier.
        """

        penalties = {}

        for unique_id, data in model.model_values.items():
            independent = data["independent"]
            dependent = data["dependent"]
            dependent_error = data.get("dependent_error", None)

            penalties[unique_id] = collate_with(
                self.penalty(
                    independent=independent,
                    dependent=dependent,
                    dependent_error=dependent_error,
                )
            )

        return penalties

    def plot_penalty(
        self,
        low_independent: Union[unyt.unyt_quantity, float],
        high_independent: Union[unyt.unyt_quantity, float],
        low_dependent: Union[unyt.unyt_quantity, float],
        high_dependent: Union[unyt.unyt_quantity, float],
        filename: str,
        x_label: Optional[str] = None,
        y_label: Optional[str] = None,
        resolution: int = 128,
        marker: Optional[str] = None,
    ):
        """
        Create a figure of the penalty function, over the limits
        given. Limits are given in log space if logged, linear
        space if not (with units required in that case).
        """

        fig, ax = plt.subplots()

        if not self.log_independent:
            low_independent = low_independent.to(self.independent_units)
            high_independent = high_independent.to(self.independent_units)

        if not self.log_dependent:
            low_dependent = low_dependent.to(self.dependent_units)
            high_dependent = high_dependent.to(self.dependent_units)

        x = np.linspace(low_independent, high_independent, resolution)
        y = np.linspace(low_dependent, high_dependent, resolution)

        xs, ys = np.meshgrid(x, y)
        output = np.empty_like(xs)

        for index, (independent, dependent) in enumerate(zip(xs.flat, ys.flat)):
            output.flat[index] = self.penalty(
                independent=np.array([independent]),
                dependent=np.array([dependent]),
                dependent_error=None,
            )

        mappable = ax.pcolormesh(xs, ys, output, vmin=0.0, vmax=1.0, rasterized=True)
        fig.colorbar(mappable=mappable, ax=ax, label="Penalty Function")

        # Plot obs data
        x = self.observation.x.to(self.independent_units)
        y = self.observation.y.to(self.dependent_units)

        if self.log_independent:
            x = np.log10(x.value)

        if self.log_dependent:
            y = np.log10(y.value)

        ax.plot(x, y, linestyle="dashed", marker=marker)

        ax.set_xlabel(x_label)
        ax.set_ylabel(y_label)

        ax.set_xlim(low_independent, high_independent)
        ax.set_ylim(low_dependent, high_dependent)

        fig.savefig(filename)

        return


@attr.s
class L1PenaltyCalculator(PenaltyCalculator):
    """
    Penalty calculator for an L1-type norm, i.e. a linear
    penalty function away from the data. This penalty function is
    capped after a (vertical) distance, provided with units
    if the provided observation is used in linear space, or
    provided as a logarithmic offset in dex if in log space.

    Parameters
    ----------

    offset: Union[unyt.unyt_quantity, float]
        The vertical offset at which to set the L1 norm to the
        maximum. This is required as the penalty function is
        not allowed to be unlimited.

    lower: Union[unyt.unyt_quantity, float]
        The lowest independent value to calculate the model
        offset at.

    upper: Union[unyt.unyt_quantity, float]
        The highest independent value to calculate the model
        offset at.
    """

    offset: Union[unyt.unyt_quantity, float] = attr.ib()
    lower: Union[unyt.unyt_quantity, float] = attr.ib()
    upper: Union[unyt.unyt_quantity, float] = attr.ib()

    def observation_interpolation(self):
        super().observation_interpolation()

        # Convert limits to sensible units and log them
        # if necessary
        if not self.log_independent:
            self.lower.convert_to_units(self.independent_units)
            self.upper.convert_to_units(self.independent_units)

        if not self.log_dependent:
            self.offset.convert_to_units(self.dependent_units)

        return

    def penalty(
        self,
        independent: np.array,
        dependent: np.array,
        dependent_error: Optional[np.array] = None,
    ) -> float:
        valid_data_mask = np.logical_and(
            independent >= self.lower, independent < self.upper
        )

        number_of_valid_points = valid_data_mask.sum()

        if number_of_valid_points == 0:
            return 0.0

        obs_dependent = self.interpolator_values(independent[valid_data_mask])

        penalties = np.abs(obs_dependent - dependent[valid_data_mask]) / self.offset

        return np.minimum(penalties, 1.0)


@attr.s
class L1PenaltyCalculatorOneSided(PenaltyCalculator):
    """
    Penalty calculator for an L1-type norm, i.e. a linear
    penalty function away from the data, but one-sided only.
    Values above/below the line are given a maximal
    penalty.

    This penalty function is capped after a (vertical) distance, provided
    with units if the provided observation is used in linear space, or
    provided as a logarithmic offset in dex if in log space.

    Parameters
    ----------

    offset: Union[unyt.unyt_quantity, float]
        The vertical offset at which to set the L1 norm to the
        maximum. This is required as the penalty function is
        not allowed to be unlimited.

    maximum_penalty: str
        Give the maximum penalty above or below the line. Accepted
        values are "above" or "below" as strings.

    lower: Union[unyt.unyt_quantity, float]
        The lowest independent value to calculate the model
        offset at.

    upper: Union[unyt.unyt_quantity, float]
        The highest independent value to calculate the model
        offset at.
    """

    offset: Union[unyt.unyt_quantity, float] = attr.ib()
    maximum_penalty: str = attr.ib()
    lower: Union[unyt.unyt_quantity, float] = attr.ib()
    upper: Union[unyt.unyt_quantity, float] = attr.ib()

    @maximum_penalty.validator
    def _check_maximum_penalty(self, attribute, value):
        valid = ["above", "below"]
        if value not in valid:
            raise AttributeError("maximum_penalty must be one of above or below.")

    def observation_interpolation(self):
        super().observation_interpolation()

        # Convert limits to sensible units and log them
        # if necessary
        if not self.log_independent:
            self.lower.convert_to_units(self.independent_units)
            self.upper.convert_to_units(self.independent_units)

        if not self.log_dependent:
            self.offset.convert_to_units(self.dependent_units)

        return

    def penalty(
        self,
        independent: np.array,
        dependent: np.array,
        dependent_error: Optional[np.array] = None,
    ) -> float:
        valid_data_mask = np.logical_and(
            independent >= self.lower, independent < self.upper
        )

        number_of_valid_points = valid_data_mask.sum()

        if number_of_valid_points == 0:
            return 0.0

        obs_dependent = self.interpolator_values(independent[valid_data_mask])

        penalties = np.abs(obs_dependent - dependent[valid_data_mask]) / self.offset

        if self.maximum_penalty == "above":
            penalties[dependent[valid_data_mask] > obs_dependent] = 1.0
        elif self.maximum_penalty == "below":
            penalties[dependent[valid_data_mask] < obs_dependent] = 1.0

        return np.minimum(penalties, 1.0)


@attr.s
class L2PenaltyCalculator(PenaltyCalculator):
    """
    Penalty calculator for an L2-type norm, i.e. a square
    penalty function away from the data. This penalty function is
    capped after a (vertical) distance, provided with units
    if the provided observation is used in linear space, or
    provided as a logarithmic offset in dex if in log space.

    Parameters
    ----------

    offset: Union[unyt.unyt_quantity, float]
        The vertical offset at which to set the L1 norm to the
        maximum. This is required as the penalty function is
        not allowed to be unlimited.

    lower: Union[unyt.unyt_quantity, float]
        The lowest independent value to calculate the model
        offset at.

    upper: Union[unyt.unyt_quantity, float]
        The highest independent value to calculate the model
        offset at.
    """

    offset: Union[unyt.unyt_quantity, float] = attr.ib()
    lower: Union[unyt.unyt_quantity, float] = attr.ib()
    upper: Union[unyt.unyt_quantity, float] = attr.ib()

    def observation_interpolation(self):
        super().observation_interpolation()

        # Convert limits to sensible units and log them
        # if necessary
        if not self.log_independent:
            self.lower.convert_to_units(self.independent_units)
            self.upper.convert_to_units(self.independent_units)

        if not self.log_dependent:
            self.offset.convert_to_units(self.dependent_units)

        return

    def penalty(
        self,
        independent: np.array,
        dependent: np.array,
        dependent_error: Optional[np.array] = None,
    ) -> float:
        valid_data_mask = np.logical_and(
            independent >= self.lower, independent < self.upper
        )

        number_of_valid_points = valid_data_mask.sum()

        if number_of_valid_points == 0:
            return 0.0

        obs_dependent = self.interpolator_values(independent[valid_data_mask])

        penalties = (obs_dependent - dependent[valid_data_mask]) ** 2 / self.offset**2

        return np.minimum(penalties, 1.0)


@attr.s
class L1VariablePenaltyCalculator(PenaltyCalculator):
    """
    Penalty calculator for an L1-type norm, i.e. a linear
    penalty function away from the data. This penalty function is
    capped after a (vertical) distance, provided with units
    if the provided observation is used in linear space, or
    provided as a logarithmic offset in dex if in log space.

    In this version the offset is variable, using a logistic
    curve.

    Parameters
    ----------

    offset_lower: Union[unyt.unyt_quantity, float]
        The vertical offset at which to set the L1 norm to the
        maximum, at the lower end of the independent range.

    offset_upper: Union[unyt.unyt_quantity, float]
        The vertical offset at which to set the L1 norm to the
        maximum, at the upper end of the independent range.

    offset_transition: Union[unyt.unyt_quantity, float]
        The independent variable at which you would like the
        offset to transition from ``offset_lower`` to
        ``offset_upper``.

    transition_width: Union[unyt.unyt_quantity, float]
        The width of the transition between offsets, centered
        around ``offset_transition``.

    lower: Union[unyt.unyt_quantity, float]
        The lowest independent value to calculate the model
        offset at.

    upper: Union[unyt.unyt_quantity, float]
        The highest independent value to calculate the model
        offset at.

    offset_below_above_ratio: float, optional
        Ratio of the allowed offset below or above the data. If this
        takes a value of less than 1.0, models below the data are penalised
        more (by that factor) than models above. Default: 1.0
    """

    offset_lower: Union[unyt.unyt_quantity, float] = attr.ib()
    offset_upper: Union[unyt.unyt_quantity, float] = attr.ib()
    offset_transition: Union[unyt.unyt_quantity, float] = attr.ib()
    transition_width: Union[unyt.unyt_quantity, float] = attr.ib()
    lower: Union[unyt.unyt_quantity, float] = attr.ib()
    upper: Union[unyt.unyt_quantity, float] = attr.ib()
    offset_below_above_ratio: float = attr.ib(default=1.0)

    def observation_interpolation(self):
        super().observation_interpolation()

        # Convert limits to sensible units and log them
        # if necessary
        if not self.log_independent:
            self.lower.convert_to_units(self.independent_units)
            self.upper.convert_to_units(self.independent_units)
            self.offset_transition.convert_to_units(self.independent_units)
            self.transition_width.convert_to_units(self.independent_units)

        if not self.log_dependent:
            self.offset_lower.convert_to_units(self.dependent_units)
            self.offset_upper.convert_to_units(self.dependent_units)

        return

    def penalty(
        self,
        independent: np.array,
        dependent: np.array,
        dependent_error: Optional[np.array] = None,
    ) -> float:
        valid_data_mask = np.logical_and(
            independent >= self.lower, independent < self.upper
        )

        number_of_valid_points = valid_data_mask.sum()

        if number_of_valid_points == 0:
            return 0.0

        offsets = self.offset_lower + (self.offset_upper - self.offset_lower) * (
            1.0
            / (
                1.0
                + np.exp(
                    (self.offset_transition - independent[valid_data_mask])
                    / self.transition_width
                )
            )
        )

        obs_dependent = self.interpolator_values(independent[valid_data_mask])
        valid_and_low_mask = dependent[valid_data_mask] < obs_dependent

        offsets[valid_and_low_mask] *= self.offset_below_above_ratio

        penalties = np.abs(obs_dependent - dependent[valid_data_mask]) / offsets

        return np.minimum(penalties, 1.0)


@attr.s
class L1SqueezePenaltyCalculator(PenaltyCalculator):
    """
    Penalty calculator for an L1-type norm, i.e. a linear
    penalty function away from the data. This penalty function is
    capped after a (vertical) distance, provided with units
    if the provided observation is used in linear space, or
    provided as a logarithmic offset in dex if in log space.

    In this version the offset is variable, with it being
    'squeezed' at a point over a width.

    Parameters
    ----------

    offset_squeeze: Union[unyt.unyt_quantity, float]
        The vertical offset at which to set the L1 norm to the
        maximum at the pinch point.

    offset_normal: Union[unyt.unyt_quantity, float]
        The usual vertical offset at which to set the L1 norm
        to the maximum.

    offset_transition: Union[unyt.unyt_quantity, float]
        The independent variable at which you would like the
        offset to transition from ``offset_lower`` to
        ``offset_upper``.

    transition_width: Union[unyt.unyt_quantity, float]
        The width of the transition between offsets, centered
        around ``offset_transition``.

    lower: Union[unyt.unyt_quantity, float]
        The lowest independent value to calculate the model
        offset at.

    upper: Union[unyt.unyt_quantity, float]
        The highest independent value to calculate the model
        offset at.

    offset_below_above_ratio: float, optional
        Ratio of the allowed offset below or above the data. If this
        takes a value of less than 1.0, models below the data are penalised
        more (by that factor) than models above. Default: 1.0
    """

    offset_squeeze: Union[unyt.unyt_quantity, float] = attr.ib()
    offset_normal: Union[unyt.unyt_quantity, float] = attr.ib()
    offset_transition: Union[unyt.unyt_quantity, float] = attr.ib()
    transition_width: Union[unyt.unyt_quantity, float] = attr.ib()
    lower: Union[unyt.unyt_quantity, float] = attr.ib()
    upper: Union[unyt.unyt_quantity, float] = attr.ib()
    offset_below_above_ratio: float = attr.ib(default=1.0)

    def observation_interpolation(self):
        super().observation_interpolation()

        # Convert limits to sensible units and log them
        # if necessary
        if not self.log_independent:
            self.lower.convert_to_units(self.independent_units)
            self.upper.convert_to_units(self.independent_units)
            self.offset_transition.convert_to_units(self.independent_units)
            self.transition_width.convert_to_units(self.independent_units)

        if not self.log_dependent:
            self.offset_squeeze.convert_to_units(self.dependent_units)
            self.offset_normal.convert_to_units(self.dependent_units)

        return

    def penalty(
        self,
        independent: np.array,
        dependent: np.array,
        dependent_error: Optional[np.array] = None,
    ) -> float:
        valid_data_mask = np.logical_and(
            independent >= self.lower, independent < self.upper
        )

        number_of_valid_points = valid_data_mask.sum()

        if number_of_valid_points == 0:
            return 0.0

        offsets = self.offset_normal - (self.offset_normal - self.offset_squeeze) * (
            np.exp(
                -0.5
                * (
                    (
                        (independent[valid_data_mask] - self.offset_transition)
                        / self.transition_width
                    )
                    ** 2
                )
            )
        )

        obs_dependent = self.interpolator_values(independent[valid_data_mask])
        valid_and_low_mask = dependent[valid_data_mask] < obs_dependent

        offsets[valid_and_low_mask] *= self.offset_below_above_ratio

        penalties = np.abs(obs_dependent - dependent[valid_data_mask]) / offsets

        return np.minimum(penalties, 1.0)


@attr.s
class GaussianDataErrorsPenaltyCalculator(PenaltyCalculator):
    """
    Penalty calculator for observations that include errors.
    This penalty function uses a Gaussian distribution around
    the data, based on the observational errors. Capped at a
    input number of sigmas away from the data.

    Parameters
    ----------

    sigma_max: Union[unyt.unyt_quantity, float]
        The number of sigmas at which the function is capped.

    lower: Union[unyt.unyt_quantity, float]
        The lowest independent value to calculate the model
        offset at.

    upper: Union[unyt.unyt_quantity, float]
        The highest independent value to calculate the model
        offset at.
    """

    sigma_max: Union[unyt.unyt_quantity, float] = attr.ib()
    lower: Union[unyt.unyt_quantity, float] = attr.ib()
    upper: Union[unyt.unyt_quantity, float] = attr.ib()

    error_interpolator_values: interp1d

    def observation_interpolation(self):
        super().observation_interpolation()

        x = self.observation.x.to(self.independent_units)
        y_scatter = self.observation.y_scatter.to(self.dependent_units)

        if self.log_independent:
            x = np.log10(x.value)

        # Propagate errors to log space if needed
        if self.log_independent:
            y = self.observation.y.to(self.dependent_units)
            y_scatter = np.abs(y_scatter / (y * np.log(10)))

        # Account for unsymmetric errors
        if y_scatter.ndim > 1:
            y_scatter = np.mean(y_scatter, axis=0)

        fill_value = lambda x, y_scatter: (y_scatter[x.argmin()], y_scatter[x.argmax()])

        self.error_interpolator_values = interp1d(
            x=x,
            y=y_scatter,
            kind="linear",
            copy=False,
            bounds_error=False,
            fill_value=fill_value(x, y_scatter),
        )

        # Convert limits to sensible units and log them
        # if necessary
        if not self.log_independent:
            self.lower.convert_to_units(self.independent_units)
            self.upper.convert_to_units(self.independent_units)

        return

    def penalty(
        self,
        independent: np.array,
        dependent: np.array,
        dependent_error: Optional[np.array] = None,
    ) -> List[float]:

        valid_data_mask = np.logical_and(
            independent >= self.lower, independent < self.upper
        )

        number_of_valid_points = valid_data_mask.sum()

        if number_of_valid_points == 0:
            return 0.0

        obs_dependent = self.interpolator_values(independent[valid_data_mask])
        obs_dependent_errors = self.error_interpolator_values(
            independent[valid_data_mask]
        )

        penalties = 1.0 - np.exp(
            -0.5
            * (
                (dependent[valid_data_mask] - obs_dependent) ** 2
                / (obs_dependent_errors**2)
            )
        )

        sigma_max_mask_value = 1.0 - np.exp(-0.5 * (self.sigma_max**2))
        penalties[penalties > sigma_max_mask_value] = 1.0

        return np.minimum(penalties, 1.0)


@attr.s
class GaussianPercentErrorsPenaltyCalculator(PenaltyCalculator):
    """
    Penalty calculator that that uses Gaussian errors with
    widths based on the percentages difference between the model
    and the data.

    Parameters
    ----------

    percent_error: float
        percent error that sets the one-sigma deviation,
        in units of percent (0-100).

    sigma_max: float
        The number of sigmas at which the function is capped.

    lower: Union[unyt.unyt_quantity, float]
        The lowest independent value to calculate the model
        offset at.

    upper: Union[unyt.unyt_quantity, float]
        The highest independent value to calculate the model
        offset at.
    """

    percent_error: float = attr.ib()
    sigma_max: float = attr.ib()
    lower: Union[unyt.unyt_quantity, float] = attr.ib()
    upper: Union[unyt.unyt_quantity, float] = attr.ib()

    def observation_interpolation(self):
        super().observation_interpolation()

        # Convert limits to sensible units and log them
        # if necessary
        if not self.log_independent:
            self.lower.convert_to_units(self.independent_units)
            self.upper.convert_to_units(self.independent_units)

        return

    def penalty(
        self,
        independent: np.array,
        dependent: np.array,
        dependent_error: Optional[np.array] = None,
    ) -> List[float]:

        valid_data_mask = np.logical_and(
            independent >= self.lower, independent < self.upper
        )

        number_of_valid_points = valid_data_mask.sum()

        if number_of_valid_points == 0:
            return 0.0

        obs_dependent = self.interpolator_values(independent[valid_data_mask])

        penalties = 1.0 - np.exp(
            -0.5
            * (
                (dependent[valid_data_mask] - obs_dependent) ** 2
                / ((obs_dependent * (self.percent_error / 100)) ** 2)
            )
        )

        sigma_max_mask_value = 1.0 - np.exp(-0.5 * (self.sigma_max**2))
        penalties[penalties > sigma_max_mask_value] = 1.0

        return np.minimum(penalties, 1.0)


@attr.s
class GaussianDataErrorsPercentFloorPenaltyCalculator(PenaltyCalculator):
    """
    Penalty calculator that that uses Gaussian errors based
    on the observational data. Includes a floor based on a
    percent error. It will pick the worst out of the two.
    This is meant as a way to not fit better then the
    emulator allows, while also not constraining stronger
    than observations.

    Parameters
    ----------

    percent_error: float
        percent error that sets the one-sigma deviation,
        in units of percent (0-100).

    sigma_max: float
        The number of sigmas at which the function is capped.
    lower: Union[unyt.unyt_quantity, float]
        The lowest independent value to calculate the model
        offset at.

    upper: Union[unyt.unyt_quantity, float]
        The highest independent value to calculate the model
        offset at.
    """

    percent_error: float = attr.ib()
    sigma_max: float = attr.ib()
    lower: Union[unyt.unyt_quantity, float] = attr.ib()
    upper: Union[unyt.unyt_quantity, float] = attr.ib()

    error_interpolator_values: interp1d

    def observation_interpolation(self):
        super().observation_interpolation()

        x = self.observation.x.to(self.independent_units)
        y_scatter = self.observation.y_scatter.to(self.dependent_units)

        if self.log_independent:
            x = np.log10(x.value)

        # Propagate errors to log space if needed
        if self.log_independent:
            y = self.observation.y.to(self.dependent_units)
            y_scatter = np.abs(y_scatter / (y * np.log(10)))

        # Account for unsymmetric errors
        if y_scatter.ndim > 1:
            y_scatter = np.mean(y_scatter, axis=0)

        fill_value = lambda x, y_scatter: (y_scatter[x.argmin()], y_scatter[x.argmax()])

        self.error_interpolator_values = interp1d(
            x=x,
            y=y_scatter,
            kind="linear",
            copy=False,
            bounds_error=False,
            fill_value=fill_value(x, y_scatter),
        )

        # Convert limits to sensible units and log them
        # if necessary
        if not self.log_independent:
            self.lower.convert_to_units(self.independent_units)
            self.upper.convert_to_units(self.independent_units)

        return

    def penalty(
        self,
        independent: np.array,
        dependent: np.array,
        dependent_error: Optional[np.array] = None,
    ) -> List[float]:

        valid_data_mask = np.logical_and(
            independent >= self.lower, independent < self.upper
        )

        number_of_valid_points = valid_data_mask.sum()

        if number_of_valid_points == 0:
            return 0.0

        obs_dependent = self.interpolator_values(independent[valid_data_mask])
        obs_dependent_errors = self.error_interpolator_values(
            independent[valid_data_mask]
        )

        penalties = 1.0 - np.exp(
            -0.5
            * (
                (dependent[valid_data_mask] - obs_dependent) ** 2
                / (
                    obs_dependent_errors**2
                    + (obs_dependent * (self.percent_error / 100)) ** 2
                )
            )
        )

        sigma_max_mask_value = 1.0 - np.exp(-0.5 * (self.sigma_max**2))
        penalties[penalties > sigma_max_mask_value] = 1.0

        return np.minimum(penalties, 1.0)


@attr.s
class GaussianWeightedDataErrorsPercentFloorPenaltyCalculator(PenaltyCalculator):
    """
    Penalty calculator that that uses Gaussian errors based
    on the observational data. Includes a floor based on a
    percent error. It will pick the worst out of the two.
    This is meant as a way to not fit better then the
    emulator allows, while also not constraining stronger
    than observations.

    Parameters
    ----------

    percent_error: float
        percent error that sets the one-sigma deviation,
        in units of percent (0-100).

    sigma_max: float
        The number of sigmas at which the function is capped.

    lower: Union[unyt.unyt_quantity, float]
        The lowest independent value to calculate the model
        offset at.

    upper: Union[unyt.unyt_quantity, float]
        The highest independent value to calculate the model
        offset at.

    weight: A general weight that scales the entire range of
        errors, but keeps relative weights intact.
    """

    percent_error: float = attr.ib()
    sigma_max: float = attr.ib()
    weight: float = attr.ib()
    lower: Union[unyt.unyt_quantity, float] = attr.ib()
    upper: Union[unyt.unyt_quantity, float] = attr.ib()

    error_interpolator_values: interp1d

    def observation_interpolation(self):
        super().observation_interpolation()

        x = self.observation.x.to(self.independent_units)
        y_scatter = self.observation.y_scatter.to(self.dependent_units)

        if self.log_independent:
            x = np.log10(x.value)

        # Propagate errors to log space if needed
        if self.log_independent:
            y = self.observation.y.to(self.dependent_units)
            y_scatter = np.abs(y_scatter / (y * np.log(10)))

        # Account for unsymmetric errors
        if y_scatter.ndim > 1:
            y_scatter = np.mean(y_scatter, axis=0)

        fill_value = lambda x, y_scatter: (y_scatter[x.argmin()], y_scatter[x.argmax()])

        self.error_interpolator_values = interp1d(
            x=x,
            y=y_scatter,
            kind="linear",
            copy=False,
            bounds_error=False,
            fill_value=fill_value(x, y_scatter),
        )

        # Convert limits to sensible units and log them
        # if necessary
        if not self.log_independent:
            self.lower.convert_to_units(self.independent_units)
            self.upper.convert_to_units(self.independent_units)

        return

    def penalty(
        self,
        independent: np.array,
        dependent: np.array,
        dependent_error: Optional[np.array] = None,
    ) -> List[float]:

        valid_data_mask = np.logical_and(
            independent >= self.lower, independent < self.upper
        )

        number_of_valid_points = valid_data_mask.sum()

        if number_of_valid_points == 0:
            return 0.0

        obs_dependent = self.interpolator_values(independent[valid_data_mask])
        obs_dependent_errors = self.error_interpolator_values(
            independent[valid_data_mask]
        )

        penalties = 1.0 - np.exp(
            -0.5
            * (
                (dependent[valid_data_mask] - obs_dependent) ** 2
                / (
                    obs_dependent_errors**2
                    + (obs_dependent * (self.percent_error / 100)) ** 2
                )
            )
        )

        sigma_max_mask_value = 1.0 - np.exp(-0.5 * (self.sigma_max**2))
        penalties[penalties > sigma_max_mask_value] = 1.0

        return np.minimum(penalties, 1.0)



"""
Fixed mean model. Basic, fixed, 0th order model.
"""

from .base import MeanModel

import numpy as np
import attr

from typing import Optional


@attr.s
class FixedMeanModel(MeanModel):
    """
    A basic offset mean model. Uses a fixed (manual) value for the mean
    model and does not allow it to be changed.
    """

    model: float = attr.ib(default=1.0)

    def train(self, independent: np.ndarray, dependent: np.ndarray) -> None:
        """
        Train the model. See :class:`MeanModel` for more information.
        """

        # This is a no-op function because the model is fixed.

        return

    def predict(self, independent: np.ndarray) -> np.ndarray:
        """
        Predict using the model. See :class:`MeanModel` for more information.
        """

        dependent = np.ones(independent.shape[0], dtype=independent.dtype) * self.model

        return dependent



"""
A basic polynomial model based on the scikit-learn multidimensional
mean model.
"""

import attr
import numpy as np
import sklearn.linear_model as lm

from sklearn.preprocessing import PolynomialFeatures
from sklearn.pipeline import Pipeline

from .base import MeanModel

from typing import Optional, Union


@attr.s
class PolynomialMeanModel(MeanModel):
    """
    A polynomial mean model; fits a linear model to the
    multidimensional parameter space.

    Under the hood, this uses the ``sklearn.linear_model.lm``.

    Parameters
    ----------

    degree: int, optional
        Maximal degree of the polynomial surface, default 1 (linear
        in each parameter).
    """

    degree: int = attr.ib(default=1)

    model: Optional[Union[lm.LinearRegression, lm.Lasso]] = None

    def train(self, independent: np.ndarray, dependent: np.ndarray) -> None:
        """
        Train the model. See :class:`MeanModel` for more information.
        """

        self.model = Pipeline(
            [
                ("poly", PolynomialFeatures(degree=self.degree)),
                ("linear", lm.LinearRegression(fit_intercept=True)),
            ]
        )

        self.model.fit(independent, dependent)

        return

    def predict(self, independent: np.ndarray) -> np.ndarray:
        """
        Predict using the model. See :class:`MeanModel` for more information.
        """

        dependent = self.model.predict(independent)

        return dependent



"""
Gaussian Process Emulator for emulating single values
"""

import attr
import copy
import numpy as np
import george

from typing import Hashable, List, Optional, Dict

from swiftemulator.emulators.base import BaseEmulator

from swiftemulator.backend.model_parameters import ModelParameters
from swiftemulator.backend.model_specification import ModelSpecification
from swiftemulator.backend.model_values import ModelValues
from swiftemulator.mean_models import MeanModel

from scipy.optimize import minimize


@attr.s
class GaussianProcessEmulator1D(BaseEmulator):
    """
    Generator for emulators for individual values. In this case
    no independent values are used. The prediction is based on
    the model parameters and model values only.

    Parameters
    ----------

    kernel, george.kernels.Kernel, optional
        The ``george`` kernel to use. The GPE here uses a copy
        of this instance. By default, this is the
        ``ExpSquaredKernel`` in George

    mean_model, MeanModel, optional
        A mean model conforming to the ``swiftemulator`` mean model
        protocol (several pre-made models are available in the
        :mod:`swiftemulator.mean_models` module).
    """

    kernel: Optional[george.kernels.Kernel] = attr.ib(default=None)
    mean_model: Optional[MeanModel] = attr.ib(default=None)

    model_specification: Optional[ModelSpecification] = None
    model_parameters: Optional[ModelParameters] = None
    model_values: Optional[ModelValues] = None

    ordering: Optional[List[Hashable]] = None
    parameter_order: Optional[List[str]] = None

    independent_variables: Optional[np.array] = None
    dependent_variables: Optional[np.array] = None
    dependent_variable_errors: Optional[np.array] = None

    emulator: Optional[george.GP] = None

    def _build_arrays(
        self,
        model_specification: ModelSpecification,
        model_parameters: ModelParameters,
        model_values: ModelValues,
    ):
        """
        Builds the arrays for passing to `george`.

        Parameters
        ----------

        model_specification: ModelSpecification
            Full instance of the model specification.

        model_parameters: ModelParameters
            Full instance of the model parameters.

        model_values: ModelValues
            Full instance of the model values describing
            this individual scaling relation.
        """

        self.model_specification = model_specification
        self.model_parameters = model_parameters
        self.model_values = model_values

        unique_identifiers = model_values.model_values.keys()
        number_of_independents = model_values.number_of_variables
        number_of_model_parameters = model_specification.number_of_parameters
        model_parameters = model_parameters.model_parameters

        independent_variables = np.empty(
            (number_of_independents, number_of_model_parameters), dtype=np.float32
        )

        dependent_variables = np.empty((number_of_independents), dtype=np.float32)

        dependent_variable_errors = np.empty((number_of_independents), dtype=np.float32)

        self.parameter_order = model_specification.parameter_names
        self.ordering = []
        filled_lines = 0

        for unique_identifier in unique_identifiers:
            self.ordering.append(unique_identifier)

            # Unpack model parameters into an array
            model_parameter_array = np.array(
                [
                    model_parameters[unique_identifier][parameter]
                    for parameter in self.parameter_order
                ]
            )

            this_model = model_values.model_values[unique_identifier]
            model_dependent = this_model["dependent"]
            model_error = this_model.get(
                "dependent_error", np.zeros(len(model_dependent))
            )

            if np.ndim(model_error) != 1:
                raise AttributeError(
                    "Multiple dimensional errors are not currently supported in GPE mode"
                )

            for line in range(len(model_dependent)):
                independent_variables[filled_lines] = model_parameter_array
                dependent_variables[filled_lines] = model_dependent[line]
                dependent_variable_errors[filled_lines] = model_error[line]

                filled_lines += 1

        assert filled_lines == number_of_independents

        self.independent_variables = independent_variables
        self.dependent_variables = dependent_variables
        self.dependent_variable_errors = dependent_variable_errors

    def fit_model(
        self,
        model_specification: ModelSpecification,
        model_parameters: ModelParameters,
        model_values: ModelValues,
    ):
        """
        Fits the gaussian process model, as determined by the
        initialiser variables of the class (i.e. the kernel and
        the mean model).

        Parameters
        ----------

        model_specification: ModelSpecification
            Full instance of the model specification.

        model_parameters: ModelParameters
            Full instance of the model parameters.

        model_values: ModelValues
            Full instance of the model values describing
            this individual scaling relation.

        Notes
        -----

        This method uses copies of the internal kernel and mean model
        objects, as those objects contain slightly unhelpful state information.
        """

        if self.independent_variables is None:
            # Creates independent_variables, dependent_variables.
            self._build_arrays(
                model_specification=model_specification,
                model_parameters=model_parameters,
                model_values=model_values,
            )

        if self.kernel is None:
            number_of_kernel_dimensions = model_specification.number_of_parameters

            self.kernel = 1**2 * george.kernels.ExpSquaredKernel(
                np.ones(number_of_kernel_dimensions), ndim=number_of_kernel_dimensions
            )

        if self.mean_model is not None:
            self.mean_model.train(
                independent=self.independent_variables,
                dependent=self.dependent_variables,
            )

            gaussian_process = george.GP(
                copy.deepcopy(self.kernel),
                fit_kernel=True,
                mean=self.mean_model.george_model,
                fit_mean=False,
            )
        else:
            gaussian_process = george.GP(
                copy.deepcopy(self.kernel),
            )

        # TODO: Figure out how to include non-symmetric errors.
        gaussian_process.compute(
            x=self.independent_variables,
            yerr=self.dependent_variable_errors,
        )

        def negative_log_likelihood(p):
            gaussian_process.set_parameter_vector(p)
            return -gaussian_process.log_likelihood(self.dependent_variables)

        def grad_negative_log_likelihood(p):
            gaussian_process.set_parameter_vector(p)
            return -gaussian_process.grad_log_likelihood(self.dependent_variables)

        # Optimize the hyperparameter values in the emulator
        result = minimize(
            fun=negative_log_likelihood,
            x0=gaussian_process.get_parameter_vector(),
            jac=grad_negative_log_likelihood,
        )

        # Load in the optimal hyperparameters
        gaussian_process.set_parameter_vector(result.x)

        self.emulator = gaussian_process

        return

    def predict_values(self, model_parameters: Dict[str, float]) -> np.array:
        """
        Predict a value and associated variance from the trained emulator contained within
        this object. Returns the value at the input model parameters.

        Parameters
        ----------

        model_parameters: Dict[str, float]
            The point in model parameter space to create predicted
            values at.

        Returns
        -------

        dependent_prediction, float
            Value of predictions, if the emulator is a function f, this
            is the predicted value of f(independent) evaluted at the position
            of the input model_parameters.

        dependent_prediction_error, float
            Error (variance) on the model prediction.

        Raises
        ------

        AttributeError
            When the model has not been trained before trying to make a
            prediction.
        """

        if self.emulator is None:
            raise AttributeError(
                "Please train the emulator with fit_model before attempting "
                "to make predictions."
            )

        model_parameter_array = np.array(
            [model_parameters[parameter] for parameter in self.parameter_order]
        )

        # Create a fake duplicate as george always needs two points to predict
        t = np.empty((2, len(model_parameter_array)), dtype=np.float32)

        t[0] = model_parameter_array
        t[1] = model_parameter_array

        model, errors = self.emulator.predict(
            y=self.dependent_variables, t=t, return_cov=False, return_var=True
        )

        return model[0], errors[0]

    def predict_values_no_error(self, model_parameters: Dict[str, float]) -> np.array:
        """
        Predict a value and associated variance from the trained emulator contained within
        this object. Returns the value at the input model parameters.

        Parameters
        ----------

        model_parameters: Dict[str, float]
            The point in model parameter space to create predicted
            values at.

        Returns
        -------

        dependent_prediction, float
            Value of predictions, if the emulator is a function f, this
            is the predicted value of f(independent) evaluted at the position
            of the input model_parameters.

        Raises
        ------

        AttributeError
            When the model has not been trained before trying to make a
            prediction.
        """

        if self.emulator is None:
            raise AttributeError(
                "Please train the emulator with fit_model before attempting "
                "to make predictions."
            )

        model_parameter_array = np.array(
            [model_parameters[parameter] for parameter in self.parameter_order]
        )

        # Create a fake duplicate as george always needs two points to predict
        t = np.empty((2, len(model_parameter_array)), dtype=np.float32)

        t[0] = model_parameter_array
        t[1] = model_parameter_array

        model = self.emulator.predict(
            y=self.dependent_variables, t=t, return_cov=False, return_var=False
        )

        return model[0]



"""
Tests model design code.
"""

from swiftemulator.backend.model_specification import ModelSpecification

from swiftemulator.design.latin import create_hypercube
from swiftemulator.design.random import create_cube


def test_create_hypercube():
    """
    Creates a hypercube and tests that values fall within the correct limits.

    Statistics testing of the hypercube is left to the individual user.
    """

    model_spec = ModelSpecification(
        number_of_parameters=4,
        parameter_names=["Hello", "World", "Foo", "Bar"],
        parameter_limits=[[-2.0, -1.0], [0.0, 1.0], [0.2, 0.3], [-1023.0, 1048.0]],
    )

    hypercube = create_hypercube(
        model_specification=model_spec,
        number_of_samples=120,
    )

    assert len(hypercube.model_parameters) == 120

    for model in hypercube.model_parameters.values():
        for parameter, limits in zip(
            model_spec.parameter_names, model_spec.parameter_limits
        ):
            sample = model[parameter]
            assert sample >= limits[0] and sample <= limits[1]

    return


def test_create_cube():
    """
    Creates a hypercube and tests that values fall within the correct limits.
    This time, we use the completely random design.

    Statistics testing of the hypercube is left to the individual user.
    """

    model_spec = ModelSpecification(
        number_of_parameters=4,
        parameter_names=["Hello", "World", "Foo", "Bar"],
        parameter_limits=[[-2.0, -1.0], [0.0, 1.0], [0.2, 0.3], [-1023.0, 1048.0]],
    )

    hypercube = create_cube(
        model_specification=model_spec,
        number_of_samples=120,
    )

    assert len(hypercube.model_parameters) == 120

    for model in hypercube.model_parameters.values():
        for parameter, limits in zip(
            model_spec.parameter_names, model_spec.parameter_limits
        ):
            sample = model[parameter]
            assert sample >= limits[0] and sample <= limits[1]

    return



"""
Basic sensitivity analysis based purely on the model values
at consistent values in the space. No emulation is used to
determine the sensitivity.

A different sensitivity analysis is ran for each dependent
variable, so it is important to ensure that the functions
are evaluated at consistent values.
"""

from swiftemulator.backend.model_parameters import ModelParameters
from swiftemulator.backend.model_values import ModelValues
from swiftemulator.backend.model_specification import ModelSpecification

from typing import Dict, Tuple, Union, Optional

import numpy as np
import matplotlib.pyplot as plt

from SALib.analyze import rbd_fast


def binwise_sensitivity(
    specification: ModelSpecification,
    parameters: ModelParameters,
    values: ModelValues,
) -> Dict[str, np.array]:
    """
    Creates a binwise sensitivity analysis dictionary.

    For each bin in dependent variable a hash is created; these
    are the keys in the returned dictionary.

    Parameters
    ----------

    specification: ModelSpecification
        Model spec; parameter limits must be valid as these feed into
        the sensitivity analysis.

    parameters: ModelParameters
        Parameters; these feed in as the independent variables in the
        sensitivity analysis.

    values: ModelValues
        Dependent variables in the sensitivity analysis.


    Returns
    -------

    sensitivity: Dict[str, np.array]
        Binwise sensitivity analysis, with each array corresponding
        to the parameters in the order as specified by the
        ``specification``. This is the "S1" vector from the
        RBD FAST method.
    """

    # Eventual structure: hash : List[List[parameters]]
    independent = {}
    # Eventual structure: hash : List[values]
    dependent = {}

    hash_function = lambda x: f"{x:1.4g}"

    for id, simulation in values.model_values.items():
        pars = [
            parameters.model_parameters[id][x] for x in specification.parameter_names
        ]
        xs = simulation["independent"]
        ys = simulation["dependent"]

        hashes = map(hash_function, xs)

        for hash, y in zip(hashes, ys):
            try:
                dependent[hash].append(y)
            except KeyError:
                dependent[hash] = [y]

            try:
                independent[hash].append(pars)
            except KeyError:
                independent[hash] = [pars]

    # These dictionaries need to now be transformed into 2D arrays
    # describing the independent and dependent variables for
    # each bin.

    independent = {k: np.array(v) for k, v in independent.items()}
    dependent = {k: np.array(v) for k, v in dependent.items()}

    sensitivities = {
        k: np.array(
            rbd_fast.analyze(
                problem=specification.salib_problem,
                X=independent[k],
                Y=dependent[k],
            )["S1"]
        )
        for k in independent.keys()
    }

    return sensitivities


def plot_binwise_sensitivity(
    specification: ModelSpecification,
    sensitivities: Dict[str, np.array],
    figure: Optional[plt.Figure] = None,
    axes: Optional[plt.Axes] = None,
    xlabel: Optional[str] = None,
    ylabel: Optional[str] = None,
    cbarlabel: Optional[str] = None,
    cmap: Optional[Union[str, plt.cm.ScalarMappable]] = None,
    vmin: float = -0.25,
    vmax: float = 0.25,
) -> Tuple[plt.Figure, plt.Axes]:
    """
    Create a figure and axis displaying the output of the sensitivity
    analysis.
    """

    if figure is None:
        figure, axes = plt.subplots(
            figsize=(
                8 * (len(sensitivities)) / 20,
                3 * specification.number_of_parameters / 7,
            )
        )

    if cmap is None:
        cmap = "RdBu_r"

    # Convert dictionary to a 2D array ready for colour mapping.
    sensitivity_map = np.zeros((len(sensitivities), specification.number_of_parameters))

    for column, values in enumerate(sensitivities.values()):
        try:
            sensitivity_map[column][:] = values[:]
        except:
            pass

    mappable = axes.imshow(
        sensitivity_map.T, vmin=vmin, vmax=vmax, cmap=cmap, origin="lower"
    )

    axes.set_yticks(range(specification.number_of_parameters))
    axes.set_yticklabels(specification.parameter_printable_names)
    axes.set_xticks(range(len(sensitivities)))
    axes.set_xticklabels(list(sensitivities.keys()), rotation=90)
    figure.colorbar(label=cbarlabel, ax=axes, mappable=mappable)
    axes.set_xlabel(xlabel)
    axes.set_ylabel(ylabel)

    return figure, axes



"""
Offset mean model. Basic 0th order model.
"""

from .base import MeanModel

import numpy as np
import attr

from typing import Optional


@attr.s
class OffsetMeanModel(MeanModel):
    """
    A basic offset mean model. Simply takes the mean
    of all of the dependent variables. Not likely to
    be useful in practice, but more of an example
    of using the protocol.
    """

    model: Optional[float] = None

    def train(self, independent: np.ndarray, dependent: np.ndarray) -> None:
        """
        Train the model. See :class:`MeanModel` for more information.
        """

        self.model = np.mean(dependent)

        return

    def predict(self, independent: np.ndarray) -> np.ndarray:
        """
        Predict using the model. See :class:`MeanModel` for more information.
        """

        dependent = np.ones(independent.shape[0], dtype=independent.dtype) * self.model

        return dependent



"""
I/O functions for reading in SWIFT simulation data.

Includes functions to read parameters to
instances of :class:``ModelParameters``, and functions
to read model data to instances of :class:``ModelValues``.

Also includes functions to write out :class:``ModelParameters``
as files, based on a base parameter file.
"""

from swiftemulator.backend.model_parameters import ModelParameters
from swiftemulator.backend.model_specification import ModelSpecification
from swiftemulator.backend.model_values import ModelValues

from typing import List, Optional, Dict, Hashable, Tuple, Union, Callable
from functools import reduce
from pathlib import Path
from math import log10

import yaml
import numpy as np


def load_pipeline_outputs(
    filenames: Dict[Hashable, Path],
    scaling_relations: List[str],
    log_independent: Optional[List[str]] = None,
    log_dependent: Optional[List[str]] = None,
    histogram_for_error: Optional[str] = None,
    sanitize_output: bool = False,
) -> Tuple[Dict[str, ModelValues], Dict[str, Dict[str, Union[str, bool]]]]:
    """
    Loads the pipeline outputs from the provided files, for
    the given specified scaling relations, into a
    :class:``ModelValues`` container.

    Parameters
    ----------

    filenames, Dict[Hashable, Path]
        Paths to files to load data from, with the keys in the
        dictionary the unique identifiers for the models to use
        throughout.

    scaling_relations: List[str]
        Top-level name for the scaling relations (i.e. the top-level
        item in the yaml file, e.g. ``stellar_mass_function_100``).

    log_independent: List[str], optional
        Scaling relations (the same as in ``scaling_relations``) where
        the independent values (given by ``centers`` in the yaml files)
        should be log-scaled (uses ``log10``).

    log_dependent: List[str], optional
        Scaling relations (the same as in ``scaling_relations``) where
        the dependent values (given by ``values`` in the yaml files)
        should be log-scaled (uses ``log10``).

    histogram_for_error: str, optional
        If provided, a histogram to use for sampling-based errors on each
        bin. This can be beneficial versus the scatter-based default errors
        when there is significant intrinsic scatter. Errors are reported
        as ``dependent / sqrt(N)`` (or ``log(1 + 1 / sqrt(N)`` in the
        case of log-scaled). Note that you _must_ ensure consistent binning
        between the other input scaling relations and this histogram, or
        there will be an error.

    sanitize_output: bool, optional
        If set to True, removes all zero-valued, infinity, and NaN dependent and
        independent values from the data. This helps sanitize input for the
        emulator, but is slower.

    Returns
    -------

    model_values, Dict[str, ModelValues]
        Dictionary of ``ModelValues`` containers for each scaling relation,
        read from the files. The keys are the names of the scaling relations.

    unit_dict, Dict[str, Dict[str, Union[str, bool]]]
        Dictionary of symbolic units for each scaling relation. Has the
        structure: ``{scaling_relation: {independent: "Msun", dependent:
        "kpc", log_independent: True, log_dependent: True}}``.

    """

    model_values = {scaling_relation: {} for scaling_relation in scaling_relations}

    unit_dict = {scaling_relation: {} for scaling_relation in scaling_relations}

    if log_independent is None:
        log_independent = []
    if log_dependent is None:
        log_dependent = []

    # Need to search for possible keys within the `lines` dictionary.
    # Priority given by ordering of line_types
    line_types = [
        "median",
        "mass_function",
        "mean",
        "adaptive_mass_function",
        "histogram",
    ]
    recursive_search = lambda d, k: (
        d.get(k[0], recursive_search(d, k[1:])) if len(k) > 0 else None
    )
    line_search = lambda d: recursive_search(d, line_types)

    for unique_identifier, filename in filenames.items():
        with open(filename, "r") as handle:
            raw_data = yaml.safe_load(handle)

        if histogram_for_error is not None:
            line = raw_data[histogram_for_error]["lines"]["histogram"]
            number_of_items = np.array(line["values"])
            inverse_square_root_error = 1.0 / np.sqrt(number_of_items)

        for scaling_relation in scaling_relations:
            line = line_search(raw_data[scaling_relation]["lines"])

            if line is None:
                continue

            independent = np.array(line["centers"])
            unit_dict[scaling_relation]["independent_units"] = line["centers_units"]

            dependent = np.array(line["values"])
            unit_dict[scaling_relation]["dependent_units"] = line["values_units"]

            dependent_error = np.array(line.get("scatter", np.zeros_like(dependent)))

            if scaling_relation in log_independent:
                independent = np.log10(independent)
                unit_dict[scaling_relation]["log_independent"] = True
            else:
                unit_dict[scaling_relation]["log_independent"] = False

            if scaling_relation in log_dependent:
                if histogram_for_error:
                    # Need to curtail at end because of bins that contain the unbinnable galaxies
                    dependent_error = np.log10(
                        1.0 + inverse_square_root_error[: len(dependent)]
                    )
                else:
                    # Handle case of dependent errors needing to be logged.
                    if dependent_error.ndim > 1:
                        lower = dependent - dependent_error[0, :]
                        upper = dependent + dependent_error[1, :]
                    else:
                        lower = dependent - dependent_error
                        upper = dependent + dependent_error

                    dependent = np.log10(dependent)
                    upper_diff = np.log10(upper) - dependent
                    lower_diff = dependent - np.log10(lower)

                    dependent_error = 0.5 * (upper_diff + lower_diff)

                unit_dict[scaling_relation]["log_dependent"] = True
            else:
                unit_dict[scaling_relation]["log_dependent"] = False

                if histogram_for_error:
                    dependent_error = (
                        dependent * inverse_square_root_error[: len(dependent)]
                    )
                else:
                    if dependent_error.ndim > 1:
                        # Only need to average if multi-dimensional, otherwise we just inherit
                        # from above.
                        dependent_error = np.mean(dependent_error, axis=0)

            if sanitize_output:
                sanitization_mask = np.logical_and.reduce(
                    [
                        np.isfinite(independent),
                        np.isfinite(dependent),
                        np.isfinite(dependent_error),
                    ]
                )
            else:
                sanitization_mask = np.s_[:]

            model_values[scaling_relation][unique_identifier] = {
                "independent": independent[sanitization_mask],
                "dependent": dependent[sanitization_mask],
                "dependent_error": dependent_error[sanitization_mask],
            }

    return {k: ModelValues(v) for k, v in model_values.items()}, unit_dict


def load_parameter_files(
    filenames: Dict[Hashable, Path],
    parameters: List[str],
    log_parameters: Optional[List[str]] = None,
    parameter_printable_names: Optional[List[str]] = None,
    parameter_limits: Optional[List[List[float]]] = None,
) -> Tuple[ModelSpecification, ModelParameters]:
    """
    Loads information from the parameter files and returns
    the associated model specification and model parameters
    instances.

    Parameters
    ----------

    filenames, Dict[Hashable, Path]
        Paths to the parameter files, keyed by their unique
        identifiers (i.e. those also used in :func:`load_pipeline_outputs`).

    parameters, List[str]
        Parameters to load from the yaml files. Should be specified
        in the same way as the ``--param`` option in SWIFT, i.e.
        in the format ``SectionName:ParameterName``.

    log_parameters, List[str], optional
        Which parameters in the list above should be scaled
        logarithmically.

    parameter_printable_names, List[str], optional
        Optional 'fancy' names for your parameters. These strings will
        be used on any figures generated through swift-emulator. Can
        include LaTeX formatting as in ``matplotlib``.

    parameter_limits, List[List[float]], optional
        The lower and upper limit of the input parameters. Should be
        the same length as ``parameters``, but each item is a list
        of length two, with a lower and upper bound. For example, in
        a two parameter model ``[[0.0, 1.0], [8.3, 9.3]]`` would mean
        that the first parameter would vary between 0.0 and 1.0, with
        the second parameter varying between 8.3 and 9.3. If not provided,
        these will be inferred from the data.

    Returns
    -------

    model_specification, ModelSpecification
        Specification for the model based on the parameters that
        have been passed to this function.

    model_parameters, ModelParameters
        Model parameter container corresponding to the SWIFT
        parameter files.
    """

    if log_parameters is None:
        log_parameters = []
    else:
        for parameter in log_parameters:
            if not parameter in parameters:
                raise AttributeError(
                    f"Parameter {parameter} requested for logarithmic transform "
                    "not available in main list of parameters."
                )

    read_parameters = {k: None for k in filenames.keys()}

    for unique_identifier, filename in filenames.items():
        with open(filename, "r") as handle:
            full_parameter_file = yaml.load(handle, Loader=yaml.Loader)

        base_parameters = {
            parameter: float(
                reduce(lambda d, k: d.get(k), parameter.split(":"), full_parameter_file)
            )
            for parameter in parameters
        }

        read_parameters[unique_identifier] = {
            parameter: log10(value) if parameter in log_parameters else value
            for parameter, value in base_parameters.items()
        }

    if parameter_limits is None:
        parameter_limits = [
            [
                min([model[parameter] for model in read_parameters.values()]),
                max([model[parameter] for model in read_parameters.values()]),
            ]
            for parameter in parameters
        ]

    model_specification = ModelSpecification(
        number_of_parameters=len(parameters),
        parameter_names=parameters,
        parameter_limits=parameter_limits,
        parameter_printable_names=parameter_printable_names,
    )

    model_parameters = ModelParameters(model_parameters=read_parameters)

    return model_specification, model_parameters


def write_parameter_files(
    filenames: Dict[Hashable, Path],
    model_parameters: ModelParameters,
    parameter_transforms: Optional[Dict[str, Callable]] = None,
    base_parameter_file: Optional[Path] = None,
):
    """
    Writes parameter files, containing the parameters from a
    :class:`ModelParameters` instance, based on a base parameter
    file.

    Parameters
    ----------

    filenames: Dict[Hashable, Path]
        Dictionary stating where to write each parameter file, based
        upon the unique identifiers of each run in ``model_parameters``.

    model_parameters: ModelParameters
        Varied parameters to write in the output files.

    parameter_transforms: Dict[str, Callable], optional
        Parameter transformation functions for transforming parameter values
        before writing. Parameters may be generated (and emulated) in a
        space that is very different to their meaning in the code. Hence,
        this parameter allows for a transformation (for instance, a logrithmic
        transformation). For each parameter that should be transformed (keys)
        there should be a function taking the emulated value, transforming it
        into the code value. For instance, if a parameter is emulated in
        logarithmic space, this should be ``lambda x: 10**x``.

    base_parameter_file: Path, optional
        Base parameter file to read. The parameters specified in
        ``model_parameters`` will be overwritten when writing each
        individual file, but the rest will remain the same.

    Notes
    -----

    Also changes the value of ``MetaData:run_name`` to the unique
    identifiers.
    """

    if base_parameter_file is not None:
        with open(base_parameter_file, "r") as handle:
            base_parameters = yaml.load(handle, Loader=yaml.Loader)
    else:
        base_parameters = {}

    if parameter_transforms is None:
        parameter_transforms = {}

    for identifier, filename in filenames.items():
        parameters = base_parameters.copy()

        for parameter, value in {
            "MetaData:run_name": str(identifier),
            **model_parameters.model_parameters[identifier],
        }.items():
            section, key = parameter.split(":")

            transform = parameter_transforms.get(parameter, lambda x: x)

            try:
                parameters[section][key] = transform(value)
            except KeyError:
                parameters[section] = {}
                parameters[section][key] = transform(value)

        with open(filename, "w") as handle:
            yaml.dump(parameters, handle, default_flow_style=False)

    return



"""
A gaussian process emulator that uses _multiple_ internal emulators to
better predict functions that contain a 'break'.
"""

import attr
import numpy as np
import george
import math

from typing import Hashable, List, Optional, Dict

from swiftemulator.emulators.base import BaseEmulator
from swiftemulator.emulators.gaussian_process import GaussianProcessEmulator
from swiftemulator.backend.model_parameters import ModelParameters
from swiftemulator.backend.model_specification import ModelSpecification
from swiftemulator.backend.model_values import ModelValues
from swiftemulator.mean_models import MeanModel

from scipy.optimize import minimize


@attr.s
class MultipleGaussianProcessEmulator(BaseEmulator):
    """
    Generator for emulators for individual scaling relations,
    using multiple trained gaussian processes regression
    instances and linear models under the hood.

    Parameters
    ----------

    kernel, george.kernels.Kernel, optional
        The ``george`` kernel to use. The GPE here uses a copy
        of this instance. By default, this is the
        ``ExpSquaredKernel`` in George

    mean_model, MeanModel, optional
        A mean model conforming to the ``swiftemulator`` mean model
        protocol (several pre-made models are available in the
        :mod:`swiftemulator.mean_models` module).

    independent_regions, List[List[float]]
        The regions over which to construct independent emulators.
        None can be used in the first and last element to specify
        there are no boundaries to overlap. Must be monotonically
        increasing. Overlaps between regions are allowed and
        predicted values will be a weighted linear combination
        of both. For example, you could use
        ``[[None, 1.0], [0.7, 2.0], [1.7, None]]`` for data that
        ran from 0.0 to 3.0 in the independent variable.
        Regions should not overlap more than once. This isn't checked,
        but will break the code.
    """

    kernel: Optional[george.kernels.Kernel] = attr.ib(default=None)
    mean_model: Optional[MeanModel] = attr.ib(default=None)
    independent_regions: Optional[List[List[float]]] = attr.ib(default=[[None, None]])

    model_specification: Optional[ModelSpecification] = None
    model_parameters: Optional[ModelParameters] = None
    model_values: Optional[ModelValues] = None
    model_values_regions: Optional[List[ModelValues]] = None

    emulators: Optional[List[george.GP]] = None

    def _build_arrays(
        self,
        model_specification: ModelSpecification,
        model_parameters: ModelParameters,
        model_values: ModelValues,
    ):
        """
        Builds the arrays for passing to sub-emulators.

        Parameters
        ----------

        model_specification: ModelSpecification
            Full instance of the model specification.

        model_parameters: ModelParameters
            Full instance of the model parameters.

        model_values: ModelValues
            Full instance of the model values describing
            this individual scaling relation.
        """

        # Here we must take our full instances and convert
        # them into separate instances only containing the sub
        # independent regions
        unique_identifiers = list(model_values.keys())

        values = [{} for _ in self.independent_regions]

        for unique_id in unique_identifiers:
            raw_values = model_values[unique_id]

            ind = raw_values["independent"]
            dep = raw_values["dependent"]

            no_error = False

            try:
                err = raw_values["dependent_error"]
            except KeyError:
                no_error = True

            for index, (low, high) in enumerate(self.independent_regions):
                mask = np.logical_and(
                    ind > low if low is not None else np.ones_like(ind).astype(bool),
                    ind < high if high is not None else np.ones_like(ind).astype(bool),
                )

                values[index][unique_id] = {
                    "independent": ind[mask],
                    "dependent": dep[mask],
                }

                if not no_error:
                    values[index][unique_id]["dependent_error"] = err[mask]

        self.model_values_regions = [ModelValues(x) for x in values]

        return

    def fit_model(
        self,
        model_specification: ModelSpecification,
        model_parameters: ModelParameters,
        model_values: ModelValues,
    ):
        """
        Fits the gaussian process model, as determined by the
        initialiser variables of the class (i.e. the kernel and
        the mean model).

        Parameters
        ----------

        model_specification: ModelSpecification
            Full instance of the model specification.

        model_parameters: ModelParameters
            Full instance of the model parameters.

        model_values: ModelValues
            Full instance of the model values describing
            this individual scaling relation.

        Notes
        -----

        This method uses copies of the internal kernel and mean model
        objects, as those objects contain slightly unhelpful state information.
        """

        if self.independent_variables is None:
            # Creates independent_variables, dependent_variables.
            self._build_arrays(
                model_specification=model_specification,
                model_parameters=model_parameters,
                model_values=model_values,
            )

        self.emulators = [
            GaussianProcessEmulator(kernel=self.kernel, mean_model=self.mean_model)
            for _ in self.independent_regions
        ]

        for gp, model_vals in zip(self.emulators, self.model_values_regions):
            gp.fit_model(
                model_specification=model_specification,
                model_parameters=model_parameters,
                model_values=model_vals,
            )

        return

    def predict_values(
        self,
        independent: np.array,
        model_parameters: Dict[str, float],
    ) -> np.array:
        """
        Predict values and the associated variance from the trained emulator contained
        within this object.

        Parameters
        ----------

        independent, np.array
            Independent continuous variables to evaluate the emulator
            at.

        model_parameters: Dict[str, float]
            The point in model parameter space to create predicted
            values at.

        Returns
        -------

        dependent_predictions, np.array
            Array of predictions, if the emulator is a function f, these
            are the predicted values of f(independent) evaluted at the position
            of the input model_parameters.

        dependent_prediction_errors, np.array
            Errors on the model predictions.

        Raises
        ------

        AttributeError
            When the model has not been trained before trying to make a
            prediction, or when attempting to evaluate the model at
            disallowed independent variables.

        Notes
        -----

        This will use the originally defined regions and overlaps will
        be calculated by using the weighted linear sum corresponding
        to the independent variable's distance to the adjacent boundary.
        The errors use a weighted square sum.
        """

        if self.emulators is None:
            raise AttributeError(
                "Please train the emulator with fit_model before attempting "
                "to make predictions."
            )

        # First, do individual predictions.

        inputs = []
        output = []
        output_error = []

        for index, (low, high) in enumerate(self.independent_regions):
            mask = np.logical_and(
                (
                    independent > low
                    if low is not None
                    else np.ones_like(independent).astype(bool)
                ),
                (
                    independent < high
                    if high is not None
                    else np.ones_like(independent).astype(bool)
                ),
            )

            predicted, errors = self.emulators[index].predict_values(
                independent=independent[mask], model_parameters=model_parameters
            )

            inputs.append(list(independent[mask]))
            output.append(list(predicted))
            output_error.append(list(errors))

        # Now that we've predicted it all, we need to explicitly deal
        # with overlap and non-overlap.

        overlap_ranges = {}

        for index in range(1, len(self.independent_regions)):
            left = self.independent_regions[index][0]
            right = self.independent_regions[index - 1][1]

            if right is None or left is None:
                continue
            elif right > left:
                overlap_ranges[index - 1] = [left, right]

        dependent_predictions = np.empty_like(independent)
        dependent_prediction_errors = np.empty_like(independent)

        current_emulator = 0

        for index, x in enumerate(independent):
            if x not in inputs[current_emulator]:
                current_emulator += 1

            # Is it in the prior overlap?
            low, high = overlap_ranges.get(current_emulator - 1, [float("inf")] * 2)

            if low <= x <= high:
                # We have already counted this independent variable.
                continue

            # Is it in this emulator's overlap?
            low, high = overlap_ranges.get(current_emulator, [float("inf")] * 2)

            if low <= x <= high:
                dependent_index_left = inputs[current_emulator].index(x)
                dependent_index_right = inputs[current_emulator + 1].index(x)

                ind_left = inputs[current_emulator][dependent_index_left]
                ind_right = inputs[current_emulator + 1][dependent_index_right]

                left_weight = (high - x) / (high - low)
                right_weight = (x - low) / (high - low)

                dependent_left = output[current_emulator][dependent_index_left]
                dependent_right = output[current_emulator + 1][dependent_index_right]

                dependent_error_left = output_error[current_emulator][
                    dependent_index_left
                ]
                dependent_error_right = output_error[current_emulator][
                    dependent_index_right
                ]

                dependent_predictions[index] = (
                    dependent_left * left_weight + dependent_right * right_weight
                )
                dependent_prediction_errors[index] = math.sqrt(
                    left_weight * dependent_error_left * dependent_error_left
                    + right_weight * dependent_error_right * dependent_error_right
                )
            else:
                # Easy!
                dependent_index = inputs[current_emulator].index(x)
                dependent_predictions[index] = output[current_emulator][dependent_index]
                dependent_prediction_errors[index] = output_error[current_emulator][
                    dependent_index
                ]

        return dependent_predictions, dependent_prediction_errors

    def predict_values_no_error(
        self,
        independent: np.array,
        model_parameters: Dict[str, float],
    ) -> np.array:
        """
        Predict values from the trained emulator contained within this object.
        In cases where the error estimates are not required, this method is
        significantly faster than predict_values().

        Parameters
        ----------

        independent, np.array
            Independent continuous variables to evaluate the emulator
            at.

        model_parameters: Dict[str, float]
            The point in model parameter space to create predicted
            values at.

        Returns
        -------

        dependent_predictions, np.array
            Array of predictions, if the emulator is a function f, these
            are the predicted values of f(independent) evaluted at the position
            of the input model_parameters.

        Raises
        ------

        AttributeError
            When the model has not been trained before trying to make a
            prediction, or when attempting to evaluate the model at
            disallowed independent variables.

        Notes
        -----

        This will use the originally defined regions and overlaps will
        be calculated by using the weighted linear sum corresponding
        to the independent variable's distance to the adjacent boundary.
        The errors use a weighted square sum.
        """

        if self.emulators is None:
            raise AttributeError(
                "Please train the emulator with fit_model before attempting "
                "to make predictions."
            )

        # First, do individual predictions.

        inputs = []
        output = []

        for index, (low, high) in enumerate(self.independent_regions):
            mask = np.logical_and(
                (
                    independent > low
                    if low is not None
                    else np.ones_like(independent).astype(bool)
                ),
                (
                    independent < high
                    if high is not None
                    else np.ones_like(independent).astype(bool)
                ),
            )

            predicted = self.emulators[index].predict_values_no_error(
                independent=independent[mask], model_parameters=model_parameters
            )

            inputs.append(list(independent[mask]))
            output.append(list(predicted))

        # Now that we've predicted it all, we need to explicitly deal
        # with overlap and non-overlap.

        overlap_ranges = {}

        for index in range(1, len(self.independent_regions)):
            left = self.independent_regions[index][0]
            right = self.independent_regions[index - 1][1]

            if right is None or left is None:
                continue
            elif right > left:
                overlap_ranges[index - 1] = [left, right]

        dependent_predictions = np.empty_like(independent)

        current_emulator = 0

        for index, x in enumerate(independent):
            if x not in inputs[current_emulator]:
                current_emulator += 1

            # Is it in the prior overlap?
            low, high = overlap_ranges.get(current_emulator - 1, [float("inf")] * 2)

            if low <= x <= high:
                # We have already counted this independent variable.
                continue

            # Is it in this emulator's overlap?
            low, high = overlap_ranges.get(current_emulator, [float("inf")] * 2)

            if low <= x <= high:
                dependent_index_left = inputs[current_emulator].index(x)
                dependent_index_right = inputs[current_emulator + 1].index(x)

                ind_left = inputs[current_emulator][dependent_index_left]
                ind_right = inputs[current_emulator + 1][dependent_index_right]

                left_weight = (high - x) / (high - low)
                right_weight = (x - low) / (high - low)

                dependent_left = output[current_emulator][dependent_index_left]
                dependent_right = output[current_emulator + 1][dependent_index_right]

                dependent_predictions[index] = (
                    dependent_left * left_weight + dependent_right * right_weight
                )
                dependent_prediction_errors[index] = math.sqrt(
                    left_weight * dependent_error_left * dependent_error_left
                    + right_weight * dependent_error_right * dependent_error_right
                )
            else:
                # Easy!
                dependent_index = inputs[current_emulator].index(x)
                dependent_predictions[index] = output[current_emulator][dependent_index]

        return dependent_predictions



"""
Basic test for any emulator. Leave one simulation out
and test how well the emulator fits those values.
"""

import attr
import numpy as np
import matplotlib.pyplot as plt
import george
import copy

from tqdm import tqdm
from pathlib import Path

from swiftemulator.backend.model_parameters import ModelParameters
from swiftemulator.backend.model_values import ModelValues
from swiftemulator.backend.model_specification import ModelSpecification

from swiftemulator.emulators.gaussian_process import GaussianProcessEmulator
from swiftemulator.emulators.gaussian_process_mcmc import GaussianProcessEmulatorMCMC
from swiftemulator.emulators.linear_model import LinearModelEmulator

from swiftemulator.mean_models import MeanModel


from typing import List, Dict, Tuple, Union, Optional, Hashable


@attr.s
class CrossCheck(object):
    """
    Generator for emulators for leave one out checks.

    Parameters
    ----------

    kernel, george.kernels
        The ``george`` kernel to use. The GPE here uses a copy
        of this instance. By default, this is the
        ``ExpSquaredKernel`` in George

    mean_model, MeanModel, optional
        A mean model conforming to the ``swiftemulator`` mean model
        protocol (several pre-made models are available in the
        :mod:`swiftemulator.mean_models` module).

    hide_progress: bool
        Option to display a tqdm bar when creating the emulators,
        Default is to hide progress bar.
    """

    kernel: Optional[george.kernels.Kernel] = attr.ib(default=None)
    mean_model: Optional[MeanModel] = attr.ib(default=None)
    hide_progress: bool = attr.ib(default=True)

    model_specification: ModelSpecification
    model_parameters: ModelParameters
    model_values: ModelValues

    leave_out_order: Optional[List[int]] = None
    cross_emulators: Optional[Dict[Hashable, george.GP]] = None

    def build_emulators(
        self,
        model_specification: ModelSpecification,
        model_parameters: ModelParameters,
        model_values: ModelValues,
    ):
        """
        Build a dictonary with an emulator for each simulation
        where the data of that simulation is left out

        Note: this can take a long time


        Parameters
        ----------

        model_specification: ModelSpecification
            Full instance of the model specification.

        model_parameters: ModelParameters
            Full instance of the model parameters.

        model_values: ModelValues
            Full instance of the model values describing
            this individual scaling relation.
        """

        self.model_specification = model_specification
        self.model_parameters = model_parameters
        self.model_values = model_values

        self.leave_out_order = list(model_values.model_values.keys())

        emulators = {}

        for unique_identifier in tqdm(self.leave_out_order, disable=self.hide_progress):
            left_out_data = model_values.model_values.pop(unique_identifier)

            emulator = GaussianProcessEmulator(
                kernel=self.kernel,
                mean_model=self.mean_model,
            )

            emulator.fit_model(
                model_specification=model_specification,
                model_parameters=model_parameters,
                model_values=model_values,
            )

            emulators[unique_identifier] = emulator

            model_values.model_values[unique_identifier] = left_out_data

        self.cross_emulators = emulators

        return

    def build_mocked_model_values_original_independent(self) -> ModelValues:
        """ "
        Builds a mocked :class:`ModelValues` container, using the cross
        emulators. The emulators are evaluated at the same independent
        variables that were 'left out'.

        Returns
        -------

        model_values: ModelValues
            The model values container with each leave-one-out
            scaling relation predicted. This is also set as
            ``cross_model_values``.
        """

        if self.cross_emulators is None:
            raise AttributeError(
                "You need to build the emulators before the prediction step."
            )

        cross_model_values = {}

        for unique_identifier in self.model_values.keys():
            independent = self.model_values[unique_identifier]["independent"]

            emulated, emulated_error = self.cross_emulators[
                unique_identifier
            ].predict_values(
                independent,
                model_parameters=self.model_parameters[unique_identifier],
            )

            cross_model_values[unique_identifier] = {
                "independent": independent,
                "dependent": emulated,
                "dependent_error": np.sqrt(emulated_error),
            }

        cross_model_values = ModelValues(cross_model_values)

        return cross_model_values

    def build_mocked_model_values(self, emulate_at: np.array) -> ModelValues:
        """
        Builds a mocked :class:`ModelValues` container, using the cross
        emulators. Similar to ``build_mocked_model_value_original_independent``
        but evaluates all emulators at a consistent set of independent
        variables.

        Parameters
        ----------

        emulate_at: np.array
            independent array where the emulator is evaluated.

        Returns
        -------

        model_values: ModelValues
            The model values container with each leave-one-out
            scaling relation predicted. This is also set as
            ``cross_model_values``.

        """

        if self.cross_emulators is None:
            raise AttributeError(
                "You need to build the emulators before the prediction step."
            )

        cross_model_values = {}

        for unique_identifier in self.model_values.keys():
            emulated, emulated_error = self.cross_emulators[
                unique_identifier
            ].predict_values(
                emulate_at,
                model_parameters=self.model_parameters[unique_identifier],
            )

            cross_model_values[unique_identifier] = {
                "independent": emulate_at,
                "dependent": emulated,
                "dependent_error": np.sqrt(emulated_error),
            }

        cross_model_values = ModelValues(cross_model_values)

        return cross_model_values

    def plot_results(
        self,
        emulate_at: np.array,
        output_path: Optional[Union[str, Path]] = None,
        xlabel: Optional[str] = None,
        ylabel: Optional[str] = None,
    ):
        """
        Make a plot of each of the leave_out emulators vs
        the original data.

        Parameters
        ----------

        emulate_at: np.array
            independent array where the emulator is evaluated.

        output_path: Union[str, Path], optional
            Optional, name of the folder where you want to save
            the figures.

        xlabel: str, optional
            Label for horizontal axis on the resultant figure.

        ylabel: str, optional
            Label for vertical axis on the resultant figure.
        """

        cross_model_values = self.build_mocked_model_values(emulate_at=emulate_at)

        for unique_identifier in self.cross_emulators.keys():
            fig, ax = plt.subplots()

            emulated = cross_model_values[unique_identifier]["dependent"]
            emulated_error = cross_model_values[unique_identifier]["dependent_error"]

            ax.fill_between(
                emulate_at,
                emulated - emulated_error,
                emulated + emulated_error,
                color="C1",
                alpha=0.3,
                linewidth=0.0,
            )

            ax.errorbar(
                self.model_values.model_values[unique_identifier]["independent"],
                self.model_values.model_values[unique_identifier]["dependent"],
                yerr=self.model_values.model_values[unique_identifier][
                    "dependent_error"
                ],
                label="True",
                marker=".",
                linestyle="none",
                color="C0",
            )

            ax.plot(emulate_at, emulated, label="Emulated", color="C1")

            ax.set_xlabel("Independent Variable" if xlabel is None else xlabel)
            ax.set_ylabel("Dependent Variable" if ylabel is None else ylabel)
            ax.legend()
            ax.set_title(f"Leave Out Run {unique_identifier}")

            if output_path is None:
                plt.show()
            else:
                fig.savefig(Path(output_path) / f"leave_out_{unique_identifier}.png")

    def get_mean_squared(
        self,
        use_dependent_error: bool = False,
        use_y_as_error: bool = False,
        use_squared_difference: bool = True,
    ):
        """
        Calculates the mean squared per simulation and the total mean squared
        of the entire set of left-out simulations.

        Parameters
        ----------

        use_dependent_error: boolean
            Use the simulation errors as weights for the mean squared calulation.
            Default is false.

        use_y_as_error: boolean
            Use the model y values as the weights for the calculation.

        use_squared_difference: boolean
            Use the simulation errors as weights for the mean squared calulation.
            Default is false.

        Returns
        -------

        total_square_mean: float
            Mean (square) error across the bins.

        mean_squared_dict: Dict[Hashable, float]
            Error per unique identifier.
        """

        mean_squared_dict = {}
        total_mean_squared = []

        for unique_identifier in self.cross_emulators.keys():
            x_model = self.model_values.model_values[unique_identifier]["independent"]
            y_model = self.model_values.model_values[unique_identifier]["dependent"]

            emulated, _ = self.cross_emulators[unique_identifier].predict_values(
                independent=x_model,
                model_parameters=self.model_parameters.model_parameters[
                    unique_identifier
                ],
            )

            if use_y_as_error:
                y_model_error = y_model
            else:
                y_model_error = self.model_values.model_values[unique_identifier][
                    "dependent_error"
                ]

            if use_dependent_error:
                uniq_mean_squared = (y_model - emulated) / y_model_error
            else:
                uniq_mean_squared = y_model - emulated

            if use_squared_difference:
                uniq_mean_squared = uniq_mean_squared**2

            mean_squared_dict[unique_identifier] = uniq_mean_squared
            total_mean_squared.extend(uniq_mean_squared)

        return np.mean(total_mean_squared), mean_squared_dict



"""
Gaussian Process Emulator
"""

import attr
import copy
import numpy as np
import george
import corner
import emcee
import matplotlib.pyplot as plt

from typing import Hashable, List, Optional, Dict

from swiftemulator.emulators.base import BaseEmulator

from swiftemulator.backend.model_parameters import ModelParameters
from swiftemulator.backend.model_specification import ModelSpecification
from swiftemulator.backend.model_values import ModelValues
from swiftemulator.mean_models import MeanModel

from scipy.optimize import minimize


@attr.s
class GaussianProcessEmulatorMCMC(BaseEmulator):
    """
    Generator for emulators for individual scaling relations.

    Parameters
    ----------

    kernel, george.kernels.Kernel, optional
        The ``george`` kernel to use. The GPE here uses a copy
        of this instance. By default, this is the
        ``ExpSquaredKernel`` in George

    mean_model, MeanModel, optional
        A mean model conforming to the ``swiftemulator`` mean model
        protocol (several pre-made models are available in the
        :mod:`swiftemulator.mean_models` module).

    burn_in_steps, int, optional
        Optional: Number of steps used for the burn-in part of the MCMC chain.
        Defaults to 50 for small intial test.

    mcmc_steps, int, optional
        Optional: Number of steps used for sampling the likelihood by the MCMC.
        chain. Defaults to 100 for small initial tests.

    walkers, int, optional
        Optional: Number of walkers used by the MCMC. Defaults to 40. Should
        (statistically) be at least 2 times the number of free parameters


    use_hyperparameter_error, bool, optional
        Switch for including errors originating from uncertain
        hyperparameters in the prediction outputs, (defaults to ``False``).

    samples_for_error, int, optional
        Number of MCMC samples to use for hyperparameter error estimation
        if ``use_hyperparameter_error`` is ``True``, defaults to 100.
    """

    kernel: Optional[george.kernels.Kernel] = attr.ib(default=None)
    mean_model: Optional[MeanModel] = attr.ib(default=None)
    burn_in_steps: int = attr.ib(default=50)
    mcmc_steps: int = attr.ib(default=100)
    walkers: int = attr.ib(default=40)
    use_hyperparameter_error: bool = attr.ib(default=False)
    samples_for_error: int = attr.ib(default=100)

    model_specification: Optional[ModelSpecification] = None
    model_parameters: Optional[ModelParameters] = None
    model_values: Optional[ModelValues] = None

    ordering: Optional[List[Hashable]] = None
    parameter_order: Optional[List[str]] = None

    independent_variables: Optional[np.array] = None
    dependent_variables: Optional[np.array] = None
    dependent_variable_errors: Optional[np.array] = None

    emulator: Optional[george.GP] = None

    def _build_arrays(
        self,
        model_specification: ModelSpecification,
        model_parameters: ModelParameters,
        model_values: ModelValues,
    ):
        """
        Builds the arrays for passing to `george`.

        Parameters
        ----------

        model_specification: ModelSpecification
            Full instance of the model specification.

        model_parameters: ModelParameters
            Full instance of the model parameters.

        model_values: ModelValues
            Full instance of the model values describing
            this individual scaling relation.
        """

        self.model_specification = model_specification
        self.model_parameters = model_parameters
        self.model_values = model_values

        unique_identifiers = model_values.model_values.keys()
        number_of_independents = model_values.number_of_variables
        number_of_model_parameters = model_specification.number_of_parameters
        model_parameters = model_parameters.model_parameters

        independent_variables = np.empty(
            (number_of_independents, number_of_model_parameters + 1), dtype=np.float32
        )

        dependent_variables = np.empty((number_of_independents), dtype=np.float32)

        dependent_variable_errors = np.empty((number_of_independents), dtype=np.float32)

        self.parameter_order = model_specification.parameter_names
        self.ordering = []
        filled_lines = 0

        for unique_identifier in unique_identifiers:
            self.ordering.append(unique_identifier)

            # Unpack model parameters into an array
            model_parameter_array = np.array(
                [
                    model_parameters[unique_identifier][parameter]
                    for parameter in self.parameter_order
                ]
            )

            this_model = model_values.model_values[unique_identifier]
            model_independent = this_model["independent"]
            model_dependent = this_model["dependent"]
            model_error = this_model.get(
                "dependent_error", np.zeros(len(model_independent))
            )

            if np.ndim(model_error) != 1:
                raise AttributeError(
                    "Multiple dimensional errors are not currently supported in GPE mode"
                )

            for line in range(len(model_independent)):
                independent_variables[filled_lines][0] = model_independent[line]
                independent_variables[filled_lines][1:] = model_parameter_array

                dependent_variables[filled_lines] = model_dependent[line]
                dependent_variable_errors[filled_lines] = model_error[line]

                filled_lines += 1

        assert filled_lines == number_of_independents

        self.independent_variables = independent_variables
        self.dependent_variables = dependent_variables
        self.dependent_variable_errors = dependent_variable_errors

    def fit_model(
        self,
        model_specification: ModelSpecification,
        model_parameters: ModelParameters,
        model_values: ModelValues,
    ):
        """
        Fits the gaussian process model, as determined by the
        initialiser variables of the class (i.e. the kernel and
        the mean model).

        Parameters
        ----------

        model_specification: ModelSpecification
            Full instance of the model specification.

        model_parameters: ModelParameters
            Full instance of the model parameters.

        model_values: ModelValues
            Full instance of the model values describing
            this individual scaling relation.

        Notes
        -----

        This method uses copies of the internal kernel and mean model
        objects, as those objects contain slightly unhelpful state information.

        """

        if self.independent_variables is None:
            # Creates independent_variables, dependent_variables.
            self._build_arrays(
                model_specification=model_specification,
                model_parameters=model_parameters,
                model_values=model_values,
            )

        if self.kernel is None:
            number_of_kernel_dimensions = (
                self.model_specification.number_of_parameters + 1
            )

            self.kernel = 1**2 * george.kernels.ExpSquaredKernel(
                np.ones(number_of_kernel_dimensions), ndim=number_of_kernel_dimensions
            )

        if self.mean_model is not None:
            self.mean_model.train(
                independent=self.independent_variables,
                dependent=self.dependent_variables,
            )

            gaussian_process = george.GP(
                copy.deepcopy(self.kernel),
                fit_kernel=True,
                mean=self.mean_model.george_model,
                fit_mean=False,
            )
        else:
            gaussian_process = george.GP(
                copy.deepcopy(self.kernel),
            )

        # TODO: Figure out how to include non-symmetric errors.
        gaussian_process.compute(
            x=self.independent_variables,
            yerr=self.dependent_variable_errors,
        )

        def negative_log_likelihood(p):
            gaussian_process.set_parameter_vector(p)
            return -gaussian_process.log_likelihood(self.dependent_variables)

        def grad_negative_log_likelihood(p):
            gaussian_process.set_parameter_vector(p)
            return -gaussian_process.grad_log_likelihood(self.dependent_variables)

        def log_likelihood(p):
            return -negative_log_likelihood(p)

        # Use scipy to find starting point to increase MCMC performance
        p0_start = minimize(
            fun=negative_log_likelihood,
            x0=gaussian_process.get_parameter_vector(),
            jac=grad_negative_log_likelihood,
        ).x

        # set up a MCMC sampling routine
        ndim = len(gaussian_process)
        sampler = emcee.EnsembleSampler(self.walkers, ndim, log_likelihood)
        # Assign starting points to the walkers, at a small distance from middle point
        p0 = p0_start + 1e-4 * np.random.randn(self.walkers, ndim)

        p0, _, _ = sampler.run_mcmc(p0, self.burn_in_steps)
        sampler.run_mcmc(p0, self.mcmc_steps)

        samples = sampler.get_chain()[:, self.burn_in_steps :, :].reshape((-1, ndim))

        result = np.mean(samples, axis=0)

        # Save the samples and best fit for error analysis
        self.hyperparameter_samples = samples
        self.hyperparameter_best_fit = result

        # Load in the optimal hyperparameters
        gaussian_process.set_parameter_vector(result)

        self.emulator = gaussian_process

        return

    def plot_hyperparameter_distribution(self, filename=None, labels=None):
        """
        Makes a cornerplot of the MCMC samples obtained when fitting the model

        Parameters
        ----------

        filename, None, str
            Name for the file to which the plot is saved. Optional, if None it
            will show the image.

        labels, None, list[Hashable]
            labels to add to the different plots. Optional, if None it will
            take the kernel names

        Note
        ----

        By using this function you solemnly swear to never try to infer anything
        from the hyperparameters, except whether they are converged.

        """

        if self.hyperparameter_samples is None:
            raise AttributeError(
                "Please train the emulator with fit_model before attempting "
                "to look at the hyperparameters."
            )

        if labels is None:
            labels = self.emulator.get_parameter_names()

        corner.corner(
            self.hyperparameter_samples,
            labels=labels,
            bins=40,
            hist_bin_factor=10,
            quantiles=[0.16, 0.5, 0.84],
            levels=(0.68, 0.95),
            color="k",
            smooth1d=1,
            smooth=True,
            max_n_ticks=5,
            plot_contours=True,
            plot_datapoints=True,
            plot_density=True,
            show_titles=True,
        )

        if filename is None:
            plt.show()
        else:
            plt.savefig(filename)

        return

    def predict_values(
        self,
        independent: np.array,
        model_parameters: Dict[str, float],
    ) -> np.array:
        """
        Predict values and the associated variance from the trained emulator contained
        within this object.

        Parameters
        ----------

        independent, np.array
            Independent continuous variables to evaluate the emulator
            at.

        model_parameters: Dict[str, float]
            The point in model parameter space to create predicted
            values at.

        Returns
        -------

        dependent_predictions, np.array
            Array of predictions, if the emulator is a function f, these
            are the predicted values of f(independent) evaluted at the position
            of the input model_parameters.

        dependent_prediction_errors, np.array
            Variance on the model predictions.

        Raises
        ------

        AttributeError
            When the model has not been trained before trying to make a
            prediction.

        ValueError
            When the number of subsamples is larger than the number of samples.
        """

        if self.emulator is None:
            raise AttributeError(
                "Please train the emulator with fit_model before attempting "
                "to make predictions."
            )

        if (
            len(self.hyperparameter_samples[:, 0]) < self.samples_for_error
            and self.use_hyperparameter_error
        ):
            raise ValueError(
                "Number of subsamples must be less then the total number of samples"
            )

        model_parameter_array = np.array(
            [model_parameters[parameter] for parameter in self.parameter_order]
        )

        t = np.empty(
            (len(independent), len(model_parameter_array) + 1), dtype=np.float32
        )

        for line, value in enumerate(independent):
            t[line][0] = value
            t[line][1:] = model_parameter_array

        model, variance = self.emulator.predict(
            y=self.dependent_variables, t=t, return_cov=False, return_var=True
        )

        if self.use_hyperparameter_error:
            # Take a subsample of the MCMC samples
            sample_indices = np.random.choice(
                range(len(self.hyperparameter_samples[:, 0])),
                self.samples_for_error,
                replace=False,
            )
            hyperparameter_error_model_array = np.empty(
                (len(independent), self.samples_for_error), dtype=np.float64
            )
            for index, sample in enumerate(sample_indices):
                self.emulator.set_parameter_vector(
                    self.hyperparameter_samples[index, :]
                )
                hyperparameter_error_model_array[:, sample] = self.emulator.predict(
                    y=self.dependent_variables, t=t, return_cov=False, return_var=False
                )

            # Calculate the variance caused by hyperparameters
            hyper_variance = np.var(hyperparameter_error_model_array, axis=1)
            self.emulator.set_parameter_vector(self.hyperparameter_best_fit)
            variance += hyper_variance

        return model, variance

    def predict_values_no_error(
        self,
        independent: np.array,
        model_parameters: Dict[str, float],
    ) -> np.array:
        """
        Predict values from the trained emulator contained within this object.
        In cases where the error estimates are not required, this method is
        significantly faster than predict_values().

        Parameters
        ----------

        independent, np.array
            Independent continuous variables to evaluate the emulator
            at.

        model_parameters: Dict[str, float]
            The point in model parameter space to create predicted
            values at.

        Returns
        -------

        dependent_predictions, np.array
            Array of predictions, if the emulator is a function f, these
            are the predicted values of f(independent) evaluted at the position
            of the input model_parameters.

        Raises
        ------

        AttributeError
            When the model has not been trained before trying to make a
            prediction.

        ValueError
            When the number of subsamples is larger than the number of samples.
        """

        if self.emulator is None:
            raise AttributeError(
                "Please train the emulator with fit_model before attempting "
                "to make predictions."
            )

        if (
            len(self.hyperparameter_samples[:, 0]) < self.samples_for_error
            and self.use_hyperparameter_error
        ):
            raise ValueError(
                "Number of subsamples must be less then the total number of samples"
            )

        model_parameter_array = np.array(
            [model_parameters[parameter] for parameter in self.parameter_order]
        )

        t = np.empty(
            (len(independent), len(model_parameter_array) + 1), dtype=np.float32
        )

        for line, value in enumerate(independent):
            t[line][0] = value
            t[line][1:] = model_parameter_array

        model = self.emulator.predict(
            y=self.dependent_variables, t=t, return_cov=False, return_var=False
        )

        return model



"""
Sub-module that uses the created emulators to re-sample the parameter
space completely, effectively creating a higher 'resolution'
(in sub-grid parameters) hypercube.
"""

import numpy as np

from swiftemulator.backend.model_specification import ModelSpecification
from swiftemulator.backend.model_values import ModelValues
from swiftemulator.backend.model_parameters import ModelParameters
from swiftemulator.design.random import create_cube

from typing import Dict, Any, Tuple, Optional


def mock_hypercube(
    emulator,
    model_specification: ModelSpecification,
    samples: int,
    predict_values_kwargs: Optional[Dict[str, Any]] = None,
) -> Tuple[ModelValues, ModelParameters]:
    """
    Create a mocked version of the cube, interpolated at random
    points using the emulator.

    Parameters
    ----------

    emulator
        An emulator object that provides a `predict_values` function.

    model_spec: ModelSpecification
        A model specification for your chosen model. The cube
        will be generated for points within the ranges specified here.

    samples: int
        Number of samples to create within your model specification.

    predict_value_kwargs: dict, optional
        Keyword arguments to pass to ``predict_values`` on the emulator
        object.

    Returns
    -------

    values: ModelValues
        Model values container with the predictions from the provided emulator
        within a new cube.

    parameters: ModelParameters
        New model parameters generated in a random cube, corresponding to
        the unique identifiers in ``values``.

    Notes
    -----

    The unique identifiers for the new simulations are prefixed with
    ``emulated_`` to prevent confusion when comparing with 'real' data.
    Samples are generated at all of the independent variable points
    present within the provided emulator's data.
    """

    predict_values_kwargs = (
        {} if predict_values_kwargs is None else predict_values_kwargs
    )

    parameters = create_cube(
        model_specification=model_specification,
        number_of_samples=samples,
        prefix_unique_id="emulated_",
    )

    independent_variables = np.unique(
        [
            item
            for uid in emulator.model_values.model_values.keys()
            for item in emulator.model_values.model_values[uid]["independent"]
        ]
    )

    emulated_models = {}

    for uid, pars in parameters.model_parameters.items():
        dep, dep_err = emulator.predict_values(
            independent=independent_variables,
            model_parameters=pars,
            **predict_values_kwargs,
        )

        emulated_models[uid] = {
            "independent": independent_variables,
            "dependent": dep,
            "dependent_error": dep_err,
        }

    return ModelValues(model_values=emulated_models), parameters


def mock_sweep(
    emulator,
    model_specification: ModelSpecification,
    samples: int,
    sweep_parameter: str,
    center_point: Dict[str, float],
    predict_values_kwargs: Optional[Dict[str, Any]] = None,
) -> Tuple[ModelValues, ModelParameters]:
    """
    Create a mocked version of the cube, interpolated evenly spaced
    along one dimension.

    Parameters
    ----------

    emulator
        An emulator object that provides a `predict_values` function.

    model_spec: ModelSpecification
        A model specification for your chosen model. The cube
        will be generated for points within the ranges specified here.

    samples: int
        Number of samples to create within your model specification.

    sweep_parameter: str
        Parameter to sweep along, from minimum to maximum in the
        ``model_spec``.

    center_point: str
        Model parameters for the center point of the sweep. The other
        model parameters will remain as these values throughout the sweep.

    predict_value_kwargs: dict, optional
        Keyword arguments to pass to ``predict_values`` on the emulator
        object.

    Returns
    -------

    values: ModelValues
        Model values container with the predictions from the provided emulator
        within a new cube.

    parameters: ModelParameters
        New model parameters generated in a sweep, corresponding to
        the unique identifiers in ``values``.

    Notes
    -----

    The unique identifiers for the new simulations are prefixed with
    ``emulated_`` to prevent confusion when comparing with 'real' data.
    Samples are generated at all of the independent variable points
    present within the provided emulator's data.
    """

    predict_values_kwargs = (
        {} if predict_values_kwargs is None else predict_values_kwargs
    )

    sweep_parameter_index = model_specification.parameter_names.index(sweep_parameter)

    sweep_values = np.linspace(
        *model_specification.parameter_limits[sweep_parameter_index], samples
    )

    parameters = ModelParameters(
        {
            f"emulated_{key}": {
                k: v if k != sweep_parameter else sweep_values[key]
                for k, v in center_point.items()
            }
            for key in range(samples)
        }
    )

    independent_variables = np.unique(
        [
            item
            for uid in emulator.model_values.model_values.keys()
            for item in emulator.model_values.model_values[uid]["independent"]
        ]
    )

    emulated_models = {}

    for uid, pars in parameters.model_parameters.items():
        dep, dep_err = emulator.predict_values(
            independent=independent_variables,
            model_parameters=pars,
            **predict_values_kwargs,
        )

        emulated_models[uid] = {
            "independent": independent_variables,
            "dependent": dep,
            "dependent_error": dep_err,
        }

    return ModelValues(model_values=emulated_models), parameters



"""
Tests for objects in ``swift-emulator/backend/model_values.py``
"""

from swiftemulator.backend.model_values import ModelValues
import pytest
import os

import numpy as np


def test_basic_model_values():
    """
    A basic test of model values.
    """

    input_model_values = {
        0: {
            "independent": np.arange(10),
            "dependent": np.random.rand(10),
            "dependent_error": np.random.rand(10),
        },
        1: {
            "independent": np.arange(10),
            "dependent": np.random.rand(10),
            "dependent_error": np.random.rand(20).reshape(10, 2),
        },
        2: {
            "independent": np.arange(10),
            "dependent": np.random.rand(10),
        },
    }

    output_model = ModelValues(model_values=input_model_values)

    assert output_model.number_of_variables == 30


def test_basic_model_values_raise():
    """
    A basic test of model values, with incorrect data
    that raises an exception.
    """

    input_model_values = {
        0: {
            "independent": np.arange(10),
            "dependent": np.random.rand(10),
            "dependent_error": np.random.rand(10),
        },
        1: {
            "independent": np.arange(10),
            "dependent": np.random.rand(10),
            "dependent_error": np.random.rand(18).reshape(9, 2),
        },
        2: {
            "independent": np.arange(10),
            "dependent": np.random.rand(10),
        },
    }

    with pytest.raises(AttributeError):
        ModelValues(model_values=input_model_values)


def test_model_value_dump_load():
    """
    Tests that the data can be written to and read from disk.
    """

    input_model_values = {
        0: {
            "independent": np.arange(10),
            "dependent": np.random.rand(10),
            "dependent_error": np.random.rand(10),
        },
        1: {
            "independent": np.arange(10),
            "dependent": np.random.rand(10),
            "dependent_error": np.random.rand(20).reshape(10, 2),
        },
        2: {
            "independent": np.arange(10),
            "dependent": np.random.rand(10),
        },
    }

    true_values = ModelValues(input_model_values)

    true_values.to_yaml("test.yml")

    read_values = ModelValues.from_yaml("test.yml")

    os.remove("test.yml")

    for uid in true_values.model_values.keys():
        for name in true_values[uid].keys():
            assert (true_values[uid][name] == read_values[uid][name]).all()



"""
Gaussian Process Emulator
"""

import attr
import copy
import numpy as np
import george

from typing import Hashable, List, Optional, Dict

from swiftemulator.emulators.base import BaseEmulator

from swiftemulator.backend.model_parameters import ModelParameters
from swiftemulator.backend.model_specification import ModelSpecification
from swiftemulator.backend.model_values import ModelValues
from swiftemulator.mean_models import MeanModel

from scipy.optimize import minimize


@attr.s
class GaussianProcessEmulator(BaseEmulator):
    """
    Generator for emulators for individual scaling relations.

    Parameters
    ----------

    kernel, george.kernels.Kernel, optional
        The ``george`` kernel to use. The GPE here uses a copy
        of this instance. By default, this is the
        ``ExpSquaredKernel`` in George

    mean_model, MeanModel, optional
        A mean model conforming to the ``swiftemulator`` mean model
        protocol (several pre-made models are available in the
        :mod:`swiftemulator.mean_models` module).
    """

    kernel: Optional[george.kernels.Kernel] = attr.ib(default=None)
    mean_model: Optional[MeanModel] = attr.ib(default=None)

    model_specification: Optional[ModelSpecification] = None
    model_parameters: Optional[ModelParameters] = None
    model_values: Optional[ModelValues] = None

    ordering: Optional[List[Hashable]] = None
    parameter_order: Optional[List[str]] = None

    independent_variables: Optional[np.array] = None
    dependent_variables: Optional[np.array] = None
    dependent_variable_errors: Optional[np.array] = None

    emulator: Optional[george.GP] = None

    def _build_arrays(
        self,
        model_specification: ModelSpecification,
        model_parameters: ModelParameters,
        model_values: ModelValues,
    ):
        """
        Builds the arrays for passing to `george`.

        Parameters
        ----------

        model_specification: ModelSpecification
            Full instance of the model specification.

        model_parameters: ModelParameters
            Full instance of the model parameters.

        model_values: ModelValues
            Full instance of the model values describing
            this individual scaling relation.
        """

        self.model_specification = model_specification
        self.model_parameters = model_parameters
        self.model_values = model_values

        unique_identifiers = model_values.model_values.keys()
        number_of_independents = model_values.number_of_variables
        number_of_model_parameters = model_specification.number_of_parameters
        model_parameters = model_parameters.model_parameters

        independent_variables = np.empty(
            (number_of_independents, number_of_model_parameters + 1), dtype=np.float32
        )

        dependent_variables = np.empty((number_of_independents), dtype=np.float32)

        dependent_variable_errors = np.empty((number_of_independents), dtype=np.float32)

        self.parameter_order = model_specification.parameter_names
        self.ordering = []
        filled_lines = 0

        for unique_identifier in unique_identifiers:
            self.ordering.append(unique_identifier)

            # Unpack model parameters into an array
            model_parameter_array = np.array(
                [
                    model_parameters[unique_identifier][parameter]
                    for parameter in self.parameter_order
                ]
            )

            this_model = model_values.model_values[unique_identifier]
            model_independent = this_model["independent"]
            model_dependent = this_model["dependent"]
            model_error = this_model.get(
                "dependent_error", np.zeros(len(model_independent))
            )

            if np.ndim(model_error) != 1:
                raise AttributeError(
                    "Multiple dimensional errors are not currently supported in GPE mode"
                )

            for line in range(len(model_independent)):
                independent_variables[filled_lines][0] = model_independent[line]
                independent_variables[filled_lines][1:] = model_parameter_array

                dependent_variables[filled_lines] = model_dependent[line]
                dependent_variable_errors[filled_lines] = model_error[line]

                filled_lines += 1

        assert filled_lines == number_of_independents

        self.independent_variables = independent_variables
        self.dependent_variables = dependent_variables
        self.dependent_variable_errors = dependent_variable_errors

    def fit_model(
        self,
        model_specification: ModelSpecification,
        model_parameters: ModelParameters,
        model_values: ModelValues,
    ):
        """
        Fits the gaussian process model, as determined by the
        initialiser variables of the class (i.e. the kernel and
        the mean model).

        Parameters
        ----------

        model_specification: ModelSpecification
            Full instance of the model specification.

        model_parameters: ModelParameters
            Full instance of the model parameters.

        model_values: ModelValues
            Full instance of the model values describing
            this individual scaling relation.

        Notes
        -----

        This method uses copies of the internal kernel and mean model
        objects, as those objects contain slightly unhelpful state information.
        """

        if self.independent_variables is None:
            # Creates independent_variables, dependent_variables.
            self._build_arrays(
                model_specification=model_specification,
                model_parameters=model_parameters,
                model_values=model_values,
            )

        if self.kernel is None:
            number_of_kernel_dimensions = model_specification.number_of_parameters + 1

            self.kernel = 1**2 * george.kernels.ExpSquaredKernel(
                np.ones(number_of_kernel_dimensions), ndim=number_of_kernel_dimensions
            )

        if self.mean_model is not None:
            self.mean_model.train(
                independent=self.independent_variables,
                dependent=self.dependent_variables,
            )

            gaussian_process = george.GP(
                copy.deepcopy(self.kernel),
                fit_kernel=True,
                mean=self.mean_model.george_model,
                fit_mean=False,
            )
        else:
            gaussian_process = george.GP(
                copy.deepcopy(self.kernel),
            )

        # TODO: Figure out how to include non-symmetric errors.
        gaussian_process.compute(
            x=self.independent_variables,
            yerr=self.dependent_variable_errors,
        )

        def negative_log_likelihood(p):
            gaussian_process.set_parameter_vector(p)
            return -gaussian_process.log_likelihood(self.dependent_variables)

        def grad_negative_log_likelihood(p):
            gaussian_process.set_parameter_vector(p)
            return -gaussian_process.grad_log_likelihood(self.dependent_variables)

        # Optimize the hyperparameter values in the emulator
        result = minimize(
            fun=negative_log_likelihood,
            x0=gaussian_process.get_parameter_vector(),
            jac=grad_negative_log_likelihood,
        )

        # Load in the optimal hyperparameters
        gaussian_process.set_parameter_vector(result.x)

        self.emulator = gaussian_process

        return

    def predict_values(
        self, independent: np.array, model_parameters: Dict[str, float]
    ) -> np.array:
        """
        Predict values and the associated variance from the trained emulator
        contained within this object.

        Parameters
        ----------

        independent, np.array
            Independent continuous variables to evaluate the emulator
            at.

        model_parameters: Dict[str, float]
            The point in model parameter space to create predicted
            values at.

        Returns
        -------

        dependent_predictions, np.array
            Array of predictions, if the emulator is a function f, these
            are the predicted values of f(independent) evaluted at the position
            of the input model_parameters.

        dependent_prediction_errors, np.array
            Errors (variances) on the model predictions.

        Raises
        ------

        AttributeError
            When the model has not been trained before trying to make a
            prediction.
        """

        if self.emulator is None:
            raise AttributeError(
                "Please train the emulator with fit_model before attempting "
                "to make predictions."
            )

        model_parameter_array = np.array(
            [model_parameters[parameter] for parameter in self.parameter_order]
        )

        t = np.empty(
            (len(independent), len(model_parameter_array) + 1), dtype=np.float32
        )

        for line, value in enumerate(independent):
            t[line][0] = value
            t[line][1:] = model_parameter_array

        model, errors = self.emulator.predict(
            y=self.dependent_variables, t=t, return_cov=False, return_var=True
        )

        return model, errors

    def predict_values_no_error(
        self, independent: np.array, model_parameters: Dict[str, float]
    ) -> np.array:
        """
        Predict values from the trained emulator contained within this object.
        In cases where the error estimates are not required, this method is
        significantly faster than predict_values().

        Parameters
        ----------

        independent, np.array
            Independent continuous variables to evaluate the emulator
            at.

        model_parameters: Dict[str, float]
            The point in model parameter space to create predicted
            values at.

        Returns
        -------

        dependent_predictions, np.array
            Array of predictions, if the emulator is a function f, these
            are the predicted values of f(independent) evaluted at the position
            of the input model_parameters.

        Raises
        ------

        AttributeError
            When the model has not been trained before trying to make a
            prediction.
        """

        if self.emulator is None:
            raise AttributeError(
                "Please train the emulator with fit_model before attempting "
                "to make predictions."
            )

        model_parameter_array = np.array(
            [model_parameters[parameter] for parameter in self.parameter_order]
        )

        t = np.empty(
            (len(independent), len(model_parameter_array) + 1), dtype=np.float32
        )

        for line, value in enumerate(independent):
            t[line][0] = value
            t[line][1:] = model_parameter_array

        model = self.emulator.predict(
            y=self.dependent_variables, t=t, return_cov=False, return_var=False
        )

        return model



"""
Basic test for any emulator. Leave one simulation out
and test how well the emulator fits those values.
"""

import attr
import numpy as np
import matplotlib.pyplot as plt
import george
import copy

from tqdm import tqdm
from pathlib import Path

from swiftemulator.backend.model_parameters import ModelParameters
from swiftemulator.backend.model_values import ModelValues
from swiftemulator.backend.model_specification import ModelSpecification
from swiftemulator.mean_models import MeanModel

from swiftemulator.emulators.gaussian_process_bins import GaussianProcessEmulatorBins

from typing import List, Dict, Tuple, Union, Optional, Hashable


@attr.s
class CrossCheckBins(object):
    """
    Generator for emulators for leave one out checks.

    Parameters
    ----------

    kernel, george.kernels
        The ``george`` kernel to use. The GPE here uses a copy
        of this instance. By default, this is the
        ``ExpSquaredKernel`` in George

    mean_model, MeanModel, optional
        A mean model conforming to the ``swiftemulator`` mean model
        protocol (several pre-made models are available in the
        :mod:`swiftemulator.mean_models` module).

    hide_progress: bool
        Option to display a tqdm bar when creating the emulators,
        Default is to hide progress bar.
    """

    kernel: Optional[george.kernels.Kernel] = attr.ib(default=None)
    mean_model: Optional[MeanModel] = attr.ib(default=None)
    hide_progress: bool = attr.ib(default=True)

    model_specification: ModelSpecification
    model_parameters: ModelParameters
    model_values: ModelValues

    leave_out_order: Optional[List[int]] = None
    cross_emulators: Optional[Dict[Hashable, george.GP]] = None

    def build_emulators(
        self,
        model_specification: ModelSpecification,
        model_parameters: ModelParameters,
        model_values: ModelValues,
    ):
        """
        Build a dictonary with an emulator for each simulation
        where the data of that simulation is left out

        Note: this can take a long time


        Parameters
        ----------

        model_specification: ModelSpecification
            Full instance of the model specification.

        model_parameters: ModelParameters
            Full instance of the model parameters.

        model_values: ModelValues
            Full instance of the model values describing
            this individual scaling relation.
        """

        self.model_specification = model_specification
        self.model_parameters = model_parameters
        self.model_values = model_values

        self.leave_out_order = list(model_values.model_values.keys())

        emulators = {}

        for unique_identifier in tqdm(self.leave_out_order, disable=self.hide_progress):
            left_out_data = model_values.model_values.pop(unique_identifier)

            emulator = GaussianProcessEmulatorBins(
                kernel=self.kernel, mean_model=self.mean_model
            )

            emulator.fit_model(
                model_specification=model_specification,
                model_parameters=model_parameters,
                model_values=model_values,
            )

            emulators[unique_identifier] = emulator

            model_values.model_values[unique_identifier] = left_out_data

        self.cross_emulators = emulators

        return

    def plot_results(
        self,
        output_path: Optional[Union[str, Path]] = None,
        xlabel: Optional[str] = None,
        ylabel: Optional[str] = None,
    ):
        """
        Make a plot of each of the leave_out emulators vs
        the original data.

        Parameters
        ----------

        output_path: Union[str, Path], optional
            Optional, name of the folder where you want to save
            the figures.

        xlabel: str, optional
            Label for horizontal axis on the resultant figure.

        ylabel: str, optional
            Label for vertical axis on the resultant figure.
        """

        for unique_identifier in self.cross_emulators.keys():
            fig, ax = plt.subplots()

            emulate_at, emulated, emulated_error = self.cross_emulators[
                unique_identifier
            ].predict_values(
                model_parameters=self.model_parameters.model_parameters[
                    unique_identifier
                ],
            )

            ax.fill_between(
                emulate_at,
                emulated - np.sqrt(emulated_error),
                emulated + np.sqrt(emulated_error),
                color="C1",
                alpha=0.3,
                linewidth=0.0,
            )

            ax.errorbar(
                self.model_values.model_values[unique_identifier]["independent"],
                self.model_values.model_values[unique_identifier]["dependent"],
                yerr=self.model_values.model_values[unique_identifier][
                    "dependent_error"
                ],
                label="True",
                marker=".",
                linestyle="none",
                color="C0",
            )

            ax.plot(emulate_at, emulated, label="Emulated", color="C1")

            ax.set_xlabel("Independent Variable" if xlabel is None else xlabel)
            ax.set_ylabel("Dependent Variable" if ylabel is None else ylabel)
            ax.legend()
            ax.set_title(f"Leave Out Run {unique_identifier}")

            if output_path is None:
                plt.show()
            else:
                fig.savefig(Path(output_path) / f"leave_out_{unique_identifier}.png")

    def get_mean_squared(
        self,
        use_dependent_error: bool = False,
        use_y_as_error: bool = False,
        use_squared_difference: bool = True,
    ):
        """
        Calculates the mean squared per simulation and the total mean squared
        of the entire set of left-out simulations.

        Parameters
        ----------

        use_dependent_error: bool
            Use the simulation errors as weights for the mean squared calculation.
            Default is false.

        use_y_as_error: boolean
            Use the model y values as the weights for the calculation.

        use_squared_difference: boolean
            Use the simulation errors as weights for the mean squared calculation.
            Default is false.

        Returns
        -------

        total_square_mean: float
            Mean (square) error across the bins.

        mean_squared_dict: Dict[Hashable, float]
            Error per unique identifier.
        """

        mean_squared_dict = {}
        total_mean_squared = []

        for unique_identifier in self.cross_emulators.keys():
            x_model = self.model_values.model_values[unique_identifier]["independent"]
            y_model = self.model_values.model_values[unique_identifier]["dependent"]

            _, emulated, _ = self.cross_emulators[unique_identifier].predict_values(
                independent=x_model,
                model_parameters=self.model_parameters.model_parameters[
                    unique_identifier
                ],
            )

            if use_y_as_error:
                y_model_error = y_model
            else:
                y_model_error = self.model_values.model_values[unique_identifier][
                    "dependent_error"
                ]

            if use_dependent_error:
                uniq_mean_squared = (y_model - emulated) / y_model_error
            else:
                uniq_mean_squared = y_model - emulated

            if use_squared_difference:
                uniq_mean_squared = uniq_mean_squared**2

            mean_squared_dict[unique_identifier] = uniq_mean_squared
            total_mean_squared.extend(uniq_mean_squared)

        return np.mean(total_mean_squared), mean_squared_dict



"""
Emulator generation object.
"""

import attr
import copy
import numpy as np

from typing import Hashable, List, Optional, Dict

from .model_parameters import ModelParameters
from .model_specification import ModelSpecification
from .model_values import ModelValues

from swiftemulator.emulators.gaussian_process import GaussianProcessEmulator
from swiftemulator.emulators.gaussian_process_mcmc import GaussianProcessEmulatorMCMC
from swiftemulator.emulators.gaussian_process_bins import GaussianProcessEmulatorBins
from swiftemulator.emulators.linear_model import LinearModelEmulator


@attr.s
class EmulatorGenerator(object):
    """
    Generator for emulators for individual scaling relations.

    Parameters
    ----------

    model_specification: ModelSpecification
        Full instance of the model specification.

    model_parameters: ModelParameters
        Full instance of the model parameters.

    Notes
    -----

    The required initialisation parameters are shared
    amongst all emulators that the emulator generator
    produces.
    """

    model_specification: ModelSpecification = attr.ib(
        validator=attr.validators.instance_of(ModelSpecification)
    )
    model_parameters: ModelParameters = attr.ib(
        validator=attr.validators.instance_of(ModelParameters)
    )

    def create_gaussian_process_emulator(
        self, model_values: ModelValues
    ) -> GaussianProcessEmulator:
        """
        Creates an individual emulator for an individual scaling
        relation described by the provided ``model_values``.

        Parameters
        ----------

        model_values, ModelValues
            The model values structure for this given scaling relation.
            This specifies the training data for the emulator.

        Returns
        -------

        emulator, GaussianProcessEmulator
            The built and trained emulator ready for prediction steps.
        """

        raise NotImplementedError(
            "This class has been deprecated. Please directly construct emualtors instead."
        )

    def create_gaussian_process_emulator_mcmc(
        self, model_values: ModelValues
    ) -> GaussianProcessEmulatorMCMC:
        """
        Creates the object needed for the hyperparameter_investigator
        function

        Parameters
        ----------

        model_values, ModelValues
            The model values structure for this given scaling relation.
            This specifies the training data for the emulator.

        Returns
        -------

        emulator, GaussianProcessEmulatorMCMC
            The built emulator ready for analysis of hyperparameters
        """

        raise NotImplementedError(
            "This class has been deprecated. Please directly construct emualtors instead."
        )

    def create_gaussian_process_emulator_bins(
        self, model_values: ModelValues
    ) -> GaussianProcessEmulatorBins:
        """
        Creates the object needed for the binned emulator

        Parameters
        ----------

        model_values, ModelValues
            The model values structure for this given scaling relation.
            This specifies the training data for the emulator.

        Returns
        -------

        emulator, GaussianProcessEmulatorMCMC
            The built emulator ready for analysis of hyperparameters
        """

        raise NotImplementedError(
            "This class has been deprecated. Please directly construct emualtors instead."
        )

    def create_linear_model_emulator(
        self, model_values: ModelValues
    ) -> LinearModelEmulator:
        """
        Creates an individual emulator for an individual scaling
        relation described by the provided ``model_values``.

        Parameters
        ----------

        model_values, ModelValues
            The model values structure for this given scaling relation.
            This specifies the training data for the emulator.

        Returns
        -------

        emulator, LinearModelEmulator
            The built and trained emulator ready for prediction steps.
        """

        raise NotImplementedError(
            "This class has been deprecated. Please directly construct emualtors instead."
        )



"""
Model parameters container, contains only the parameters
that are part of the differences between models (i.e. not
anything about the individual scaling relations!).
"""

import attr
import numpy as np
import matplotlib.pyplot as plt
import corner
import yaml
import json


from sklearn.neighbors import KDTree
from typing import Dict, Hashable, List, Tuple, Optional, Union, Any
from pathlib import Path

from swiftemulator.backend.model_specification import ModelSpecification


@attr.s
class ModelParameters(object):
    """
    Class that contains the parameters of the models, i.e.
    this does not contain any information about individual
    scaling relations. Performs validation on the dictionary
    that is given.

    Parameters
    ----------

    model_parameters, Dict[Hashable, Dict[str, float]]
        Free parameters of the underlying model. This is
        specified as a dictionary with the following
        structure: ``{unique_run_identifier: {parameter_name:
        parameter_value}}``. Here the unique run identifier
        can be anything, but it must be unique between runs.
        An example could be just an integer defining a run
        number. The parameter names must match with
        what is defined in the :class:``ModelSpecification``,
        with the parameter values the specific values taken
        for that individual simulation run. Note that all
        models must have each parameter present, and this
        is checked at the creation time of the ``ModelParameters``
        object.

    Raises
    ------

    AttributeError
        When the parameters do not match between all models.
    """

    model_parameters: Dict[Hashable, Dict[str, float]] = attr.ib()

    @model_parameters.validator
    def _check_model_parameters(self, attribute, value):
        # Basic validation - make sure we have the same
        # parameters in each dictionary.

        # Get an _example set_ of parameters.
        example_parameters = set(next(iter(value.values())).keys())

        for parameter_set in value.values():
            if not example_parameters == set(parameter_set.keys()):
                raise AttributeError(
                    "Models do not all have the same set of parameters."
                )

    def items(self):
        return self.model_parameters.items()

    def keys(self):
        return self.model_parameters.keys()

    def values(self):
        return self.model_parameters.values()

    def __getitem__(self, key):
        return self.model_parameters[key]

    def __len__(self):
        return len(self.model_parameters)

    def find_closest_model(
        self, comparison_parameters: Dict[str, float], number_of_close_models: int = 1
    ) -> Tuple[List[Hashable], List[Dict[str, float]]]:
        """
        Finds the closest model currently in this instance of
        ``ModelParameters`` to the set of provided
        ``comparison_parameters``, with the option to return
        the closest 'n' sets to the input.

        Parameters
        ----------

        comparison_parameters, Dict[str, float]
            Set of comparison parameters. The closest parameters,
            and unique indentifier, of the run within the current
            set of ``model_parameters`` to this point in
            n-dimensional parameter space will be returned.

        number_of_close_models, int
            Number of closest model that will be returned


        Returns
        -------

        unique_identifier, List[Hashable]
            Unique identifier of the closest run(s).

        closest_parameters, List[Dict[str, float]]
            Model parameters of the closest run(s).
        """

        # Convert the parameter dictionary to a numpy array
        parameter_ordering = list(comparison_parameters.keys())
        parameter_array = []
        model_ordering = []

        for identifier, model in self.model_parameters.items():
            model_ordering.append(identifier)
            parameter_array.append(
                [model[parameter] for parameter in parameter_ordering]
            )

        parameter_array = np.array(parameter_array)

        tree = KDTree(parameter_array, leaf_size=2)
        # Convert the closest point to a numpy array to use for the tree
        search_point = np.array(
            [comparison_parameters[parameter] for parameter in parameter_ordering]
        )

        ind = tree.query(
            search_point.reshape(1, -1),
            k=number_of_close_models,
            return_distance=False,
        )
        # Return in the correct format
        closest_models = [model_ordering[i] for i in ind[0]]
        closest_parameters = [self.model_parameters[i] for i in closest_models]

        return closest_models, closest_parameters

    def plot_model(
        self,
        model_specification: ModelSpecification,
        filename: Optional[Union[Path, str]] = None,
        corner_kwargs: Optional[Dict[str, Any]] = None,
    ):
        """
        Plots the model parameters based on the model specification
        given. Can either be saved to file, or show.

        Parameters
        ----------

        model_specification, ModelSpecification
            Model specification object for this set of parameters.

        filename, Union[str, Path], optional
            Name for the file to which the plot is saved. Optional, if None it
            will show the image.

        corner_kwargs: Dict[str, Any], optional
            Optional key word arguments to pass to `corner` for
            the plotting.

        """

        # Convert internal data to a format that corner can accept

        corner_data = np.empty(
            (len(self.model_parameters), model_specification.number_of_parameters),
            dtype=np.float32,
        )

        for index, model in enumerate(self.model_parameters.values()):
            corner_data[index] = np.array(
                [model[parameter] for parameter in model_specification.parameter_names],
                dtype=np.float32,
            )

        if corner_kwargs is None:
            corner_kwargs = {}

        corner.corner(
            corner_data,
            labels=model_specification.parameter_printable_names,
            range=model_specification.parameter_limits,
            plot_contours=False,
            plot_datapoints=True,
            plot_density=False,
            data_kwargs={"alpha": 1.0},
            **corner_kwargs,
        )

        if filename is None:
            plt.show()
        else:
            plt.savefig(filename)

    def to_yaml(self, filename: Path):
        """
        Write the model parameters to a YAML file.

        Parameters
        ----------

        filename: Path
            The path to write the file to. This should be a `Path` object,
            but if it is a string it will be automatically converted.

        """

        filename = Path(filename)

        with open(filename, "w") as handle:
            yaml.dump(self.model_parameters, stream=handle)

        return

    @classmethod
    def from_yaml(cls, filename: Path) -> "ModelParameters":
        """
        Generate an instance of :class:`ModelParameters` from a YAML file,
        written to disk using ``to_yaml``.

        Parameters
        ----------

        filename: Path
            The path to read the file from. This should be a ``Path`` object,
            but if it is a string it will be automatically converted.

        Returns
        -------

        model_parameters: ModelParameters
            Instance of ``ModelParameters`` restored from disk.

        """

        filename = Path(filename)

        with open(filename, "r") as handle:
            raw_values = dict(yaml.load(stream=handle, Loader=yaml.FullLoader))

        return cls(model_parameters=raw_values)

    def to_json(self, filename: Path):
        """
        Write the model parameters to a JSON file. Preferred over YAML
        as this is much faster for large datasets.

        Parameters
        ----------

        filename: Path
            The path to write the file to. This should be a `Path` object,
            but if it is a string it will be automatically converted.

        """

        filename = Path(filename)

        with open(filename, "w") as handle:
            json.dump(self.model_parameters, handle)

        return

    @classmethod
    def from_json(cls, filename: Path) -> "ModelParameters":
        """
        Generate an instance of :class:`ModelParameters` from a JSON file,
        written to disk using ``to_json``.

        Parameters
        ----------

        filename: Path
            The path to read the file from. This should be a ``Path`` object,
            but if it is a string it will be automatically converted.

        Returns
        -------

        model_parameters: ModelParameters
            Instance of ``ModelParameters`` restored from disk.

        """

        filename = Path(filename)

        with open(filename, "r") as handle:
            raw_values = dict(json.load(handle))

        return cls(model_parameters=raw_values)



"""
A base model to be overloaded in further implementations.
"""

from george.modeling import CallableModel
from copy import deepcopy

import numpy as np
import attr


@attr.s
class MeanModel(object):
    """
    A base mean model which describes the basic layout.
    Each function must be overloaded.
    """

    def train(self, independent: np.ndarray, dependent: np.ndarray) -> None:
        """
        Trains the mean model to predict dependent(independent).

        By convention, if there is an underlying model object,
        this should be stored in ``self.model``.

        Parameters
        ----------

        independent: np.ndarray
            Independent variables. Should be in the same format as is passed
            to ``george``.

        dependent: np.ndarray
            Dependent variables, to be predicted from the independent
            variables. Should be in the same format as is passed to  ``george``.

        Raises
        ------

        NotImplementedError
            If not implemented.
        """

        raise NotImplementedError

    def predict(self, independent: np.ndarray) -> np.ndarray:
        """
        Predicts dependent variables from independent variables,
        using the predictive model.

        Parameters
        ----------

        independent: np.ndarray
            Independent variables to predict dependent variables from,
            in the same format as is passed to ``george``.


        Returns
        -------

        dependent: np.ndarray
            Dependent variables that are in the same format that
            ``george`` expects.


        Raises
        ------

        NotImplementedError
            If not implemented

        AttributeError
            If the ``model`` is not trained.

        """

        raise NotImplementedError

    @property
    def george_model(self) -> CallableModel:
        """
        Get the ``george`` ``CallableModel`` instance that
        corresponds to this system. Returns a copy of the predict
        function associated with the current instance.
        """

        return CallableModel(self.copy().predict)

    def copy(self) -> "MeanModel":
        """
        Copy self to a new version of the model. Required should
        you wish to re-use a version of this object later on, as
        otherwise the ``model`` parameter will be mutated.
        """

        return deepcopy(self)



"""
Tests the plotting and sensitivity analysis functions.
"""

from swiftemulator.sensitivity.basic import (
    binwise_sensitivity,
    plot_binwise_sensitivity,
)
from swiftemulator.backend.emulator_generator import (
    ModelParameters,
    ModelSpecification,
    ModelValues,
)

from matplotlib.pyplot import close

import numpy as np


def test_basic_binwise_sens():
    """
    Basic binwise sensitivity test; does not confirm functionality,
    as that would be too difficult in this scenario. This function
    just tests that given a set-up, we can construct and plot
    a sensitivity analysis without crashing.
    """

    model_spec = ModelSpecification(
        number_of_parameters=2,
        parameter_names=["x", "y"],
        parameter_limits=[[1.0, 3.0], [0.0, 1.0]],
    )

    my_model_parameters = {
        0: {
            "x": 1,
            "y": 1,
        },
        1: {
            "x": 2,
            "y": 1,
        },
        2: {"x": 3, "y": 1},
    }

    model_parameters = ModelParameters(model_parameters=my_model_parameters)

    input_model_values = {
        0: {
            "independent": np.arange(10),
            "dependent": np.random.rand(10),
            "dependent_error": np.random.rand(10),
        },
        1: {
            "independent": np.arange(10),
            "dependent": np.random.rand(10),
            "dependent_error": np.random.rand(10),
        },
        2: {
            "independent": np.arange(10),
            "dependent": np.random.rand(10),
        },
    }

    model_values = ModelValues(model_values=input_model_values)

    binwise = binwise_sensitivity(
        specification=model_spec,
        parameters=model_parameters,
        values=model_values,
    )

    fig, ax = plot_binwise_sensitivity(specification=model_spec, sensitivities=binwise)

    close(fig)
